"""
The top level interface used to translate configuration data back to the
correct cloud modules
"""

import copy
import glob
import logging
import multiprocessing
import os
import signal
import time
import traceback
from itertools import groupby

import salt.client
import salt.config
import salt.loader
import salt.syspaths
import salt.utils.args
import salt.utils.cloud
import salt.utils.context
import salt.utils.crypt
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.files
import salt.utils.user
import salt.utils.verify
import salt.utils.yaml
from salt.exceptions import (
    SaltCloudConfigError,
    SaltCloudException,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)
from salt.template import compile_template

log = logging.getLogger(__name__)


def communicator(func):
    """Warning, this is a picklable decorator !"""

    def _call(queue, args, kwargs):
        """called with [queue, args, kwargs] as first optional arg"""
        kwargs["queue"] = queue
        ret = None
        try:
            ret = func(*args, **kwargs)
            queue.put("END")
        except KeyboardInterrupt as ex:
            trace = traceback.format_exc()
            queue.put("KEYBOARDINT")
            queue.put("Keyboard interrupt")
            queue.put(f"{ex}\n{trace}\n")
        except Exception as ex:  # pylint: disable=broad-except
            trace = traceback.format_exc()
            queue.put("ERROR")
            queue.put("Exception")
            queue.put(f"{ex}\n{trace}\n")
        except SystemExit as ex:
            trace = traceback.format_exc()
            queue.put("ERROR")
            queue.put("System exit")
            queue.put(f"{ex}\n{trace}\n")
        return ret

    return _call


def enter_mainloop(
    target,
    mapped_args=None,
    args=None,
    kwargs=None,
    pool=None,
    pool_size=None,
    callback=None,
    queue=None,
):
    """
    Manage a multiprocessing pool

    - If the queue does not output anything, the pool runs indefinitely

    - If the queue returns KEYBOARDINT or ERROR, this will kill the pool
      totally calling terminate & join and ands with a SaltCloudSystemExit
      exception notifying callers from the abnormal termination

    - If the queue returns END or callback is defined and returns True,
      it just join the process and return the data.

    target
        the function you want to execute in multiprocessing
    pool
        pool object can be None if you want a default pool, but you ll
        have then to define pool_size instead
    pool_size
        pool size if you did not provide yourself a pool
    callback
        a boolean taking a string in argument which returns True to
        signal that 'target' is finished and we need to join
        the pool
    queue
        A custom multiprocessing queue in case you want to do
        extra stuff and need it later in your program
    args
        positional arguments to call the function with
        if you don't want to use pool.map

    mapped_args
        a list of one or more arguments combinations to call the function with
        e.g. (foo, [[1], [2]]) will call::

                foo([1])
                foo([2])

    kwargs
        kwargs to give to the function in case of process

    Attention, the function must have the following signature:

            target(queue, *args, **kw)

    You may use the 'communicator' decorator to generate such a function
    (see end of this file)
    """
    if not kwargs:
        kwargs = {}
    if not pool_size:
        pool_size = 1
    if not pool:
        pool = multiprocessing.Pool(pool_size)
    if not queue:
        manager = multiprocessing.Manager()
        queue = manager.Queue()

    if mapped_args is not None and not mapped_args:
        msg = (
            "We are called to asynchronously execute {}"
            " but we do no have anything to execute, weird,"
            " we bail out".format(target)
        )
        log.error(msg)
        raise SaltCloudSystemExit(f"Exception caught\n{msg}")
    elif mapped_args is not None:
        iterable = [[queue, [arg], kwargs] for arg in mapped_args]
        ret = pool.map(func=target, iterable=iterable)
    else:
        ret = pool.apply(target, [queue, args, kwargs])
    while True:
        test = queue.get()
        if test in ["ERROR", "KEYBOARDINT"]:
            type_ = queue.get()
            trace = queue.get()
            msg = f"Caught {type_}, terminating workers\n"
            msg += f"TRACE: {trace}\n"
            log.error(msg)
            pool.terminate()
            pool.join()
            raise SaltCloudSystemExit(f"Exception caught\n{msg}")
        elif test in ["END"] or (callback and callback(test)):
            pool.close()
            pool.join()
            break
        else:
            time.sleep(0.125)
    return ret


class CloudClient:
    """
    The client class to wrap cloud interactions
    """

    def __init__(self, path=None, opts=None, config_dir=None, pillars=None):
        if opts:
            self.opts = opts
        else:
            self.opts = salt.config.cloud_config(path)

        # Check the cache-dir exists. If not, create it.
        v_dirs = [self.opts["cachedir"]]
        salt.utils.verify.verify_env(v_dirs, salt.utils.user.get_user())

        if pillars:
            for name, provider in pillars.pop("providers", {}).items():
                driver = provider["driver"]
                provider["profiles"] = {}
                self.opts["providers"].update({name: {driver: provider}})
            for name, profile in pillars.pop("profiles", {}).items():
                provider = profile["provider"].split(":")[0]
                driver = next(iter(self.opts["providers"][provider].keys()))
                profile["provider"] = f"{provider}:{driver}"
                profile["profile"] = name
                self.opts["profiles"].update({name: profile})
                self.opts["providers"][provider][driver]["profiles"].update(
                    {name: profile}
                )
            for name, map_dct in pillars.pop("maps", {}).items():
                if "maps" not in self.opts:
                    self.opts["maps"] = {}
                self.opts["maps"][name] = map_dct
            self.opts.update(pillars)

    def _opts_defaults(self, **kwargs):
        """
        Set the opts dict to defaults and allow for opts to be overridden in
        the kwargs
        """
        # Let's start with the default salt cloud configuration
        opts = salt.config.DEFAULT_CLOUD_OPTS.copy()
        # Update it with the loaded configuration
        opts.update(self.opts.copy())
        # Reset some of the settings to sane values
        opts["parallel"] = False
        opts["keep_tmp"] = False
        opts["deploy"] = True
        opts["update_bootstrap"] = False
        opts["show_deploy_args"] = False
        opts["script_args"] = ""
        # Update it with the passed kwargs
        if "kwargs" in kwargs:
            opts.update(kwargs["kwargs"])
        opts.update(kwargs)
        profile = opts.get("profile", None)
        # filter other profiles if one is specified
        if profile:
            tmp_profiles = opts.get("profiles", {}).copy()
            for _profile in [a for a in tmp_profiles]:
                if not _profile == profile:
                    tmp_profiles.pop(_profile)
            # if profile is specified and we have enough info about providers
            # also filter them to speedup methods like
            # __filter_non_working_providers
            providers = [
                a.get("provider", "").split(":")[0]
                for a in tmp_profiles.values()
                if a.get("provider", "")
            ]
            if providers:
                _providers = opts.get("providers", {})
                for provider in _providers.copy():
                    if provider not in providers:
                        _providers.pop(provider)
        return opts

    def low(self, fun, low):
        """
        Pass the cloud function and low data structure to run
        """
        l_fun = getattr(self, fun)
        f_call = salt.utils.args.format_call(l_fun, low)
        return l_fun(*f_call.get("args", ()), **f_call.get("kwargs", {}))

    def list_sizes(self, provider=None):
        """
        List all available sizes in configured cloud systems
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        return salt.utils.data.simple_types_filter(mapper.size_list(provider))

    def list_images(self, provider=None):
        """
        List all available images in configured cloud systems
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        return salt.utils.data.simple_types_filter(mapper.image_list(provider))

    def list_locations(self, provider=None):
        """
        List all available locations in configured cloud systems
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        return salt.utils.data.simple_types_filter(mapper.location_list(provider))

    def query(self, query_type="list_nodes"):
        """
        Query basic instance information
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        mapper.opts["selected_query_option"] = "list_nodes"
        return mapper.map_providers_parallel(query_type)

    def full_query(self, query_type="list_nodes_full"):
        """
        Query all instance information
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        mapper.opts["selected_query_option"] = "list_nodes_full"
        return mapper.map_providers_parallel(query_type)

    def select_query(self, query_type="list_nodes_select"):
        """
        Query select instance information
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        mapper.opts["selected_query_option"] = "list_nodes_select"
        return mapper.map_providers_parallel(query_type)

    def min_query(self, query_type="list_nodes_min"):
        """
        Query select instance information
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        mapper.opts["selected_query_option"] = "list_nodes_min"
        return mapper.map_providers_parallel(query_type)

    def profile(self, profile, names, vm_overrides=None, **kwargs):
        """
        Pass in a profile to create, names is a list of vm names to allocate

            vm_overrides is a special dict that will be per node options
            overrides

        Example:

        .. code-block:: python

            >>> client= salt.cloud.CloudClient(path='/etc/salt/cloud')
            >>> client.profile('do_512_git', names=['minion01',])
            {'minion01': {'backups_active': 'False',
                    'created_at': '2014-09-04T18:10:15Z',
                    'droplet': {'event_id': 31000502,
                                 'id': 2530006,
                                 'image_id': 5140006,
                                 'name': 'minion01',
                                 'size_id': 66},
                    'id': '2530006',
                    'image_id': '5140006',
                    'ip_address': '107.XXX.XXX.XXX',
                    'locked': 'True',
                    'name': 'minion01',
                    'private_ip_address': None,
                    'region_id': '4',
                    'size_id': '66',
                    'status': 'new'}}


        """
        if not vm_overrides:
            vm_overrides = {}
        kwargs["profile"] = profile
        mapper = salt.cloud.Map(self._opts_defaults(**kwargs))
        if isinstance(names, str):
            names = names.split(",")
        return salt.utils.data.simple_types_filter(
            mapper.run_profile(profile, names, vm_overrides=vm_overrides)
        )

    def map_run(self, path=None, **kwargs):
        """
        To execute a map
        """
        kwarg = {}
        if path:
            kwarg["map"] = path
        kwarg.update(kwargs)
        mapper = salt.cloud.Map(self._opts_defaults(**kwarg))
        dmap = mapper.map_data()
        return salt.utils.data.simple_types_filter(mapper.run_map(dmap))

    def destroy(self, names):
        """
        Destroy the named VMs
        """
        mapper = salt.cloud.Map(self._opts_defaults(destroy=True))
        if isinstance(names, str):
            names = names.split(",")
        return salt.utils.data.simple_types_filter(mapper.destroy(names))

    def create(self, provider, names, **kwargs):
        """
        Create the named VMs, without using a profile

        Example:

        .. code-block:: python

            client.create(provider='my-ec2-config', names=['myinstance'],
                image='ami-1624987f', size='t1.micro', ssh_username='ec2-user',
                securitygroup='default', delvol_on_destroy=True)
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        providers = self.opts["providers"]
        if provider in providers:
            provider += f":{next(iter(providers[provider].keys()))}"
        else:
            return False
        if isinstance(names, str):
            names = names.split(",")
        ret = {}
        for name in names:
            vm_ = kwargs.copy()
            vm_["name"] = name
            vm_["driver"] = provider

            # This function doesn't require a profile, but many cloud drivers
            # check for profile information (which includes the provider key) to
            # help with config file debugging and setting up instances. Setting
            # the profile and provider defaults here avoids errors in other
            # cloud functions relying on these keys. See SaltStack Issue #41971
            # and PR #38166 for more information.
            vm_["profile"] = None
            vm_["provider"] = provider

            ret[name] = salt.utils.data.simple_types_filter(mapper.create(vm_))
        return ret

    def extra_action(self, names, provider, action, **kwargs):
        """
        Perform actions with block storage devices

        Example:

        .. code-block:: python

            client.extra_action(names=['myblock'], action='volume_create',
                provider='my-nova', kwargs={'voltype': 'SSD', 'size': 1000}
            )
            client.extra_action(names=['salt-net'], action='network_create',
                provider='my-nova', kwargs={'cidr': '192.168.100.0/24'}
            )
        """
        mapper = salt.cloud.Map(self._opts_defaults())
        providers = mapper.map_providers_parallel()
        if provider in providers:
            provider += f":{next(iter(providers[provider].keys()))}"
        else:
            return False
        if isinstance(names, str):
            names = names.split(",")

        ret = {}
        for name in names:
            extra_ = kwargs.copy()
            extra_["name"] = name
            extra_["provider"] = provider
            extra_["profile"] = None
            extra_["action"] = action
            ret[name] = salt.utils.data.simple_types_filter(mapper.extras(extra_))
        return ret

    def action(
        self,
        fun=None,
        cloudmap=None,
        names=None,
        provider=None,
        instance=None,
        kwargs=None,
    ):
        """
        Execute a single action via the cloud plugin backend

        Examples:

        .. code-block:: python

            client.action(fun='show_instance', names=['myinstance'])
            client.action(fun='show_image', provider='my-ec2-config',
                kwargs={'image': 'ami-10314d79'}
            )
        """
        if kwargs is None:
            kwargs = {}

        mapper = salt.cloud.Map(self._opts_defaults(action=fun, names=names, **kwargs))
        if instance:
            if names:
                raise SaltCloudConfigError(
                    "Please specify either a list of 'names' or a single "
                    "'instance', but not both."
                )
            names = [instance]

        if names and not provider:
            self.opts["action"] = fun
            return mapper.do_action(names, kwargs)

        if provider and not names:
            return mapper.do_function(provider, fun, kwargs)
        else:
            # This should not be called without either an instance or a
            # provider. If both an instance/list of names and a provider
            # are given, then we also need to exit. We can only have one
            # or the other.
            raise SaltCloudConfigError(
                "Either an instance (or list of names) or a provider must be "
                "specified, but not both."
            )


class Cloud:
    """
    An object for the creation of new VMs
    """

    def __init__(self, opts):
        self.opts = opts
        self.clouds = salt.loader.clouds(self.opts)
        self.__filter_non_working_providers()
        self.__cached_provider_queries = {}

    def get_configured_providers(self):
        """
        Return the configured providers
        """
        providers = set()
        for alias, drivers in self.opts["providers"].items():
            if len(drivers) > 1:
                for driver in drivers:
                    providers.add(f"{alias}:{driver}")
                continue
            providers.add(alias)
        return providers

    def lookup_providers(self, lookup):
        """
        Get a dict describing the configured providers
        """
        if lookup is None:
            lookup = "all"
        if lookup == "all":
            providers = set()
            for alias, drivers in self.opts["providers"].items():
                for driver in drivers:
                    providers.add((alias, driver))

            if not providers:
                raise SaltCloudSystemExit("There are no cloud providers configured.")

            return providers

        if ":" in lookup:
            alias, driver = lookup.split(":")
            if (
                alias not in self.opts["providers"]
                or driver not in self.opts["providers"][alias]
            ):
                raise SaltCloudSystemExit(
                    "No cloud providers matched '{}'. Available: {}".format(
                        lookup, ", ".join(self.get_configured_providers())
                    )
                )

        providers = set()
        for alias, drivers in self.opts["providers"].items():
            for driver in drivers:
                if lookup in (alias, driver):
                    providers.add((alias, driver))

        if not providers:
            raise SaltCloudSystemExit(
                "No cloud providers matched '{}'. Available selections: {}".format(
                    lookup, ", ".join(self.get_configured_providers())
                )
            )
        return providers

    def lookup_profiles(self, provider, lookup):
        """
        Return a dictionary describing the configured profiles
        """
        if provider is None:
            provider = "all"
        if lookup is None:
            lookup = "all"

        if lookup == "all":
            profiles = set()
            provider_profiles = set()
            for alias, info in self.opts["profiles"].items():
                providers = info.get("provider")

                if providers:
                    given_prov_name = providers.split(":")[0]
                    salt_prov_name = providers.split(":")[1]
                    if given_prov_name == provider:
                        provider_profiles.add((alias, given_prov_name))
                    elif salt_prov_name == provider:
                        provider_profiles.add((alias, salt_prov_name))
                    profiles.add((alias, given_prov_name))

            if not profiles:
                raise SaltCloudSystemExit("There are no cloud profiles configured.")

            if provider != "all":
                return provider_profiles

            return profiles

    def map_providers(self, query="list_nodes", cached=False):
        """
        Return a mapping of what named VMs are running on what VM providers
        based on what providers are defined in the configuration and VMs
        """
        if cached is True and query in self.__cached_provider_queries:
            return self.__cached_provider_queries[query]

        pmap = {}
        for alias, drivers in self.opts["providers"].items():
            for driver, details in drivers.items():
                fun = f"{driver}.{query}"
                if fun not in self.clouds:
                    log.error("Public cloud provider %s is not available", driver)
                    continue
                if alias not in pmap:
                    pmap[alias] = {}

                try:
                    with salt.utils.context.func_globals_inject(
                        self.clouds[fun],
                        __active_provider_name__=":".join([alias, driver]),
                    ):
                        pmap[alias][driver] = self.clouds[fun]()
                except Exception as err:  # pylint: disable=broad-except
                    log.debug(
                        "Failed to execute '%s()' while querying for running nodes: %s",
                        fun,
                        err,
                        exc_info_on_loglevel=logging.DEBUG,
                    )
                    # Failed to communicate with the provider, don't list any
                    # nodes
                    pmap[alias][driver] = []
        self.__cached_provider_queries[query] = pmap
        return pmap

    def map_providers_parallel(self, query="list_nodes", cached=False):
        """
        Return a mapping of what named VMs are running on what VM providers
        based on what providers are defined in the configuration and VMs

        Same as map_providers but query in parallel.
        """
        if cached is True and query in self.__cached_provider_queries:
            return self.__cached_provider_queries[query]

        opts = self.opts.copy()
        multiprocessing_data = []

        # Optimize Providers
        opts["providers"] = self._optimize_providers(opts["providers"])
        for alias, drivers in opts["providers"].items():
            # Make temp query for this driver to avoid overwrite next
            this_query = query
            for driver, details in drivers.items():
                # If driver has function list_nodes_min, just replace it
                # with query param to check existing vms on this driver
                # for minimum information, Otherwise still use query param.
                if (
                    opts.get("selected_query_option") is None
                    and f"{driver}.list_nodes_min" in self.clouds
                ):
                    this_query = "list_nodes_min"

                fun = f"{driver}.{this_query}"
                if fun not in self.clouds:
                    log.error("Public cloud provider %s is not available", driver)
                    continue

                multiprocessing_data.append(
                    {
                        "fun": fun,
                        "opts": opts,
                        "query": this_query,
                        "alias": alias,
                        "driver": driver,
                    }
                )
        output = {}
        if not multiprocessing_data:
            return output

        data_count = len(multiprocessing_data)
        pool = multiprocessing.Pool(
            data_count < 10 and data_count or 10, init_pool_worker
        )
        parallel_pmap = enter_mainloop(
            _run_parallel_map_providers_query, multiprocessing_data, pool=pool
        )
        for alias, driver, details in parallel_pmap:
            if not details:
                # There's no providers details?! Skip it!
                continue
            if alias not in output:
                output[alias] = {}
            output[alias][driver] = details

        self.__cached_provider_queries[query] = output
        return output

    def get_running_by_names(
        self, names, query="list_nodes", cached=False, profile=None
    ):
        if isinstance(names, str):
            names = [names]

        matches = {}
        handled_drivers = {}
        mapped_providers = self.map_providers_parallel(query, cached=cached)
        for alias, drivers in mapped_providers.items():
            for driver, vms in drivers.items():
                if driver not in handled_drivers:
                    handled_drivers[driver] = alias
                # When a profile is specified, only return an instance
                # that matches the provider specified in the profile.
                # This solves the issues when many providers return the
                # same instance. For example there may be one provider for
                # each availability zone in amazon in the same region, but
                # the search returns the same instance for each provider
                # because amazon returns all instances in a region, not
                # availability zone.
                if (
                    profile
                    and alias
                    not in self.opts["profiles"][profile]["provider"].split(":")[0]
                ):
                    continue

                for vm_name, details in vms.items():
                    # XXX: The logic below can be removed once the aws driver
                    # is removed
                    if vm_name not in names:
                        continue

                    elif (
                        driver == "ec2"
                        and "aws" in handled_drivers
                        and "aws" in matches[handled_drivers["aws"]]
                        and vm_name in matches[handled_drivers["aws"]]["aws"]
                    ):
                        continue
                    elif (
                        driver == "aws"
                        and "ec2" in handled_drivers
                        and "ec2" in matches[handled_drivers["ec2"]]
                        and vm_name in matches[handled_drivers["ec2"]]["ec2"]
                    ):
                        continue

                    if alias not in matches:
                        matches[alias] = {}
                    if driver not in matches[alias]:
                        matches[alias][driver] = {}
                    matches[alias][driver][vm_name] = details

        return matches

    def _optimize_providers(self, providers):
        """
        Return an optimized mapping of available providers
        """
        new_providers = {}
        provider_by_driver = {}

        for alias, driver in providers.items():
            for name, data in driver.items():
                if name not in provider_by_driver:
                    provider_by_driver[name] = {}

                provider_by_driver[name][alias] = data

        for driver, providers_data in provider_by_driver.items():
            fun = f"{driver}.optimize_providers"
            if fun not in self.clouds:
                log.debug("The '%s' cloud driver is unable to be optimized.", driver)

                for name, prov_data in providers_data.items():
                    if name not in new_providers:
                        new_providers[name] = {}
                    new_providers[name][driver] = prov_data
                continue

            new_data = self.clouds[fun](providers_data)
            if new_data:
                for name, prov_data in new_data.items():
                    if name not in new_providers:
                        new_providers[name] = {}
                    new_providers[name][driver] = prov_data

        return new_providers

    def location_list(self, lookup="all"):
        """
        Return a mapping of all location data for available providers
        """
        data = {}

        lookups = self.lookup_providers(lookup)
        if not lookups:
            return data

        for alias, driver in lookups:
            fun = f"{driver}.avail_locations"
            if fun not in self.clouds:
                # The capability to gather locations is not supported by this
                # cloud module
                log.debug(
                    "The '%s' cloud driver defined under '%s' provider "
                    "alias is unable to get the locations information",
                    driver,
                    alias,
                )
                continue

            if alias not in data:
                data[alias] = {}

            try:

                with salt.utils.context.func_globals_inject(
                    self.clouds[fun], __active_provider_name__=":".join([alias, driver])
                ):
                    data[alias][driver] = self.clouds[fun]()
            except Exception as err:  # pylint: disable=broad-except
                log.error(
                    "Failed to get the output of '%s()': %s",
                    fun,
                    err,
                    exc_info_on_loglevel=logging.DEBUG,
                )
        return data

    def image_list(self, lookup="all"):
        """
        Return a mapping of all image data for available providers
        """
        data = {}

        lookups = self.lookup_providers(lookup)
        if not lookups:
            return data

        for alias, driver in lookups:
            fun = f"{driver}.avail_images"
            if fun not in self.clouds:
                # The capability to gather images is not supported by this
                # cloud module
                log.debug(
                    "The '%s' cloud driver defined under '%s' provider "
                    "alias is unable to get the images information",
                    driver,
                    alias,
                )
                continue

            if alias not in data:
                data[alias] = {}

            try:
                with salt.utils.context.func_globals_inject(
                    self.clouds[fun], __active_provider_name__=":".join([alias, driver])
                ):
                    data[alias][driver] = self.clouds[fun]()
            except Exception as err:  # pylint: disable=broad-except
                log.error(
                    "Failed to get the output of '%s()': %s",
                    fun,
                    err,
                    exc_info_on_loglevel=logging.DEBUG,
                )
        return data

    def size_list(self, lookup="all"):
        """
        Return a mapping of all image data for available providers
        """
        data = {}

        lookups = self.lookup_providers(lookup)
        if not lookups:
            return data

        for alias, driver in lookups:
            fun = f"{driver}.avail_sizes"
            if fun not in self.clouds:
                # The capability to gather sizes is not supported by this
                # cloud module
                log.debug(
                    "The '%s' cloud driver defined under '%s' provider "
                    "alias is unable to get the sizes information",
                    driver,
                    alias,
                )
                continue

            if alias not in data:
                data[alias] = {}

            try:
                with salt.utils.context.func_globals_inject(
                    self.clouds[fun], __active_provider_name__=":".join([alias, driver])
                ):
                    data[alias][driver] = self.clouds[fun]()
            except Exception as err:  # pylint: disable=broad-except
                log.error(
                    "Failed to get the output of '%s()': %s",
                    fun,
                    err,
                    exc_info_on_loglevel=logging.DEBUG,
                )
        return data

    def provider_list(self, lookup="all"):
        """
        Return a mapping of all image data for available providers
        """
        data = {}
        lookups = self.lookup_providers(lookup)
        if not lookups:
            return data

        for alias, driver in lookups:
            if alias not in data:
                data[alias] = {}
            if driver not in data[alias]:
                data[alias][driver] = {}
        return data

    def profile_list(self, provider, lookup="all"):
        """
        Return a mapping of all configured profiles
        """
        data = {}
        lookups = self.lookup_profiles(provider, lookup)

        if not lookups:
            return data

        for alias, driver in lookups:
            if alias not in data:
                data[alias] = {}
            if driver not in data[alias]:
                data[alias][driver] = {}
        return data

    def create_all(self):
        """
        Create/Verify the VMs in the VM data
        """
        ret = []

        for vm_name, vm_details in self.opts["profiles"].items():
            ret.append({vm_name: self.create(vm_details)})

        return ret

    def destroy(self, names, cached=False):
        """
        Destroy the named VMs
        """
        processed = {}
        names = set(names)
        matching = self.get_running_by_names(names, cached=cached)
        vms_to_destroy = set()
        parallel_data = []
        for alias, drivers in matching.items():
            for driver, vms in drivers.items():
                for name in vms:
                    if name in names:
                        vms_to_destroy.add((alias, driver, name))
                        if self.opts["parallel"]:
                            parallel_data.append(
                                {
                                    "opts": self.opts,
                                    "name": name,
                                    "alias": alias,
                                    "driver": driver,
                                }
                            )

        # destroying in parallel
        if self.opts["parallel"] and parallel_data:
            # set the pool size based on configuration or default to
            # the number of machines we're destroying
            if "pool_size" in self.opts:
                pool_size = self.opts["pool_size"]
            else:
                pool_size = len(parallel_data)
            log.info("Destroying in parallel mode; Cloud pool size: %s", pool_size)

            # kick off the parallel destroy
            output_multip = enter_mainloop(
                _destroy_multiprocessing, parallel_data, pool_size=pool_size
            )

            # massage the multiprocessing output a bit
            ret_multip = {}
            for obj in output_multip:
                ret_multip.update(obj)

            # build up a data structure similar to what the non-parallel
            # destroy uses
            for obj in parallel_data:
                alias = obj["alias"]
                driver = obj["driver"]
                name = obj["name"]
                if alias not in processed:
                    processed[alias] = {}
                if driver not in processed[alias]:
                    processed[alias][driver] = {}
                processed[alias][driver][name] = ret_multip[name]
                if name in names:
                    names.remove(name)

        # not destroying in parallel
        else:
            log.info("Destroying in non-parallel mode.")
            for alias, driver, name in vms_to_destroy:
                fun = f"{driver}.destroy"
                with salt.utils.context.func_globals_inject(
                    self.clouds[fun], __active_provider_name__=":".join([alias, driver])
                ):
                    ret = self.clouds[fun](name)
                if alias not in processed:
                    processed[alias] = {}
                if driver not in processed[alias]:
                    processed[alias][driver] = {}
                processed[alias][driver][name] = ret
                if name in names:
                    names.remove(name)

        # now the processed data structure contains the output from either
        # the parallel or non-parallel destroy and we should finish up
        # with removing minion keys if necessary
        for alias, driver, name in vms_to_destroy:
            ret = processed[alias][driver][name]
            if not ret:
                continue

            vm_ = {
                "name": name,
                "profile": None,
                "provider": ":".join([alias, driver]),
                "driver": driver,
            }
            minion_dict = salt.config.get_cloud_config_value(
                "minion", vm_, self.opts, default={}
            )
            key_file = os.path.join(
                self.opts["pki_dir"], "minions", minion_dict.get("id", name)
            )
            globbed_key_file = glob.glob(f"{key_file}.*")

            if not os.path.isfile(key_file) and not globbed_key_file:
                # There's no such key file!? It might have been renamed
                if isinstance(ret, dict) and "newname" in ret:
                    salt.utils.cloud.remove_key(self.opts["pki_dir"], ret["newname"])
                continue

            if os.path.isfile(key_file) and not globbed_key_file:
                # Single key entry. Remove it!
                salt.utils.cloud.remove_key(
                    self.opts["pki_dir"], os.path.basename(key_file)
                )
                continue

            # Since we have globbed matches, there are probably some keys for which their minion
            # configuration has append_domain set.
            if (
                not os.path.isfile(key_file)
                and globbed_key_file
                and len(globbed_key_file) == 1
            ):
                # Single entry, let's remove it!
                salt.utils.cloud.remove_key(
                    self.opts["pki_dir"], os.path.basename(globbed_key_file[0])
                )
                continue

            # Since we can't get the profile or map entry used to create
            # the VM, we can't also get the append_domain setting.
            # And if we reached this point, we have several minion keys
            # who's name starts with the machine name we're deleting.
            # We need to ask one by one!?
            print(
                "There are several minion keys who's name starts "
                "with '{}'. We need to ask you which one should be "
                "deleted:".format(name)
            )
            while True:
                for idx, filename in enumerate(globbed_key_file):
                    print(f" {idx}: {os.path.basename(filename)}")
                selection = input("Which minion key should be deleted(number)? ")
                try:
                    selection = int(selection)
                except ValueError:
                    print(f"'{selection}' is not a valid selection.")

                try:
                    filename = os.path.basename(globbed_key_file.pop(selection))
                except Exception:  # pylint: disable=broad-except
                    continue

                delete = input(f"Delete '{filename}'? [Y/n]? ")
                if delete == "" or delete.lower().startswith("y"):
                    salt.utils.cloud.remove_key(self.opts["pki_dir"], filename)
                    print(f"Deleted '{filename}'")
                    break

                print(f"Did not delete '{filename}'")
                break

        if names and not processed:
            # These machines were asked to be destroyed but could not be found
            raise SaltCloudSystemExit(
                "The following VM's were not found: {}".format(", ".join(names))
            )

        elif names and processed:
            processed["Not Found"] = names

        elif not processed:
            raise SaltCloudSystemExit("No machines were destroyed!")

        return processed

    def reboot(self, names):
        """
        Reboot the named VMs
        """
        ret = []
        pmap = self.map_providers_parallel()
        acts = {}
        for prov, nodes in pmap.items():
            acts[prov] = []
            for node in nodes:
                if node in names:
                    acts[prov].append(node)
        for prov, names_ in acts.items():
            fun = f"{prov}.reboot"
            for name in names_:
                ret.append({name: self.clouds[fun](name)})

        return ret

    def create(self, vm_, local_master=True, sync_sleep=3):
        """
        Create a single VM
        """
        output = {}

        minion_dict = salt.config.get_cloud_config_value(
            "minion", vm_, self.opts, default={}
        )

        alias, driver = vm_["provider"].split(":")
        fun = f"{driver}.create"
        if fun not in self.clouds:
            log.error(
                "Creating '%s' using '%s' as the provider "
                "cannot complete since '%s' is not available",
                vm_["name"],
                vm_["provider"],
                driver,
            )
            return

        deploy = salt.config.get_cloud_config_value("deploy", vm_, self.opts)
        make_master = salt.config.get_cloud_config_value("make_master", vm_, self.opts)

        if deploy:
            if not make_master and "master" not in minion_dict:
                log.warning(
                    "There's no master defined on the '%s' VM settings.", vm_["name"]
                )

            if "pub_key" not in vm_ and "priv_key" not in vm_:
                log.debug("Generating minion keys for '%s'", vm_["name"])
                priv, pub = salt.utils.cloud.gen_keys(
                    salt.config.get_cloud_config_value("keysize", vm_, self.opts)
                )
                vm_["pub_key"] = pub
                vm_["priv_key"] = priv
        else:
            # Note(pabelanger): We still reference pub_key and priv_key when
            # deploy is disabled.
            vm_["pub_key"] = None
            vm_["priv_key"] = None

        key_id = minion_dict.get("id", vm_["name"])

        domain = vm_.get("domain")
        if vm_.get("use_fqdn") and domain:
            minion_dict["append_domain"] = domain

        if "append_domain" in minion_dict:
            key_id = ".".join([key_id, minion_dict["append_domain"]])

        if make_master is True and "master_pub" not in vm_ and "master_pem" not in vm_:
            log.debug("Generating the master keys for '%s'", vm_["name"])
            master_priv, master_pub = salt.utils.cloud.gen_keys(
                salt.config.get_cloud_config_value("keysize", vm_, self.opts)
            )
            vm_["master_pub"] = master_pub
            vm_["master_pem"] = master_priv

        if local_master is True and deploy is True:
            # Accept the key on the local master
            salt.utils.cloud.accept_key(self.opts["pki_dir"], vm_["pub_key"], key_id)

        vm_["os"] = salt.config.get_cloud_config_value("script", vm_, self.opts)

        try:
            vm_["inline_script"] = salt.config.get_cloud_config_value(
                "inline_script", vm_, self.opts
            )
        except KeyError:
            pass

        try:
            alias, driver = vm_["provider"].split(":")
            func = f"{driver}.create"
            with salt.utils.context.func_globals_inject(
                self.clouds[fun], __active_provider_name__=":".join([alias, driver])
            ):
                output = self.clouds[func](vm_)
            if output is not False and "sync_after_install" in self.opts:
                if self.opts["sync_after_install"] and self.opts[
                    "sync_after_install"
                ] not in (
                    "all",
                    "beacons",
                    "clouds",
                    "engines",
                    "executors",
                    "grains",
                    "log",
                    "matchers",
                    "modules",
                    "output",
                    "pillar",
                    "proxymodules",
                    "renderers",
                    "returners",
                    "sdb",
                    "serializers",
                    "states",
                    "thorium",
                    "utils",
                ):
                    log.warning(
                        "Bad option for sync_after_install. Defaulting to 'all'"
                    )
                    self.opts["sync_after_install"] = "all"

                # A small pause helps the sync work more reliably
                time.sleep(sync_sleep)

                expiration_time = time.time() + 60
                while time.time() < expiration_time:
                    # We'll try every <timeout> seconds, up to a minute
                    mopts_ = copy.deepcopy(salt.config.DEFAULT_MASTER_OPTS)
                    conf_path = "/".join(self.opts["conf_file"].split("/")[:-1])
                    mopts_.update(
                        salt.config.master_config(os.path.join(conf_path, "master"))
                    )

                    with salt.client.get_local_client(mopts=mopts_) as client:
                        ret = client.cmd(
                            vm_["name"],
                            "saltutil.sync_{}".format(self.opts["sync_after_install"]),
                            timeout=self.opts["timeout"],
                        )
                        if ret:
                            log.info(
                                "Synchronized the following dynamic modules: %s", ret
                            )
                            break
        except KeyError as exc:
            log.exception(
                "Failed to create VM %s. Configuration value %s needs to be set",
                vm_["name"],
                exc,
            )
        # If it's a map then we need to respect the 'requires'
        # so we do it later
        try:
            opt_map = self.opts["map"]
        except KeyError:
            opt_map = False
        if self.opts["parallel"] and self.opts["start_action"] and not opt_map:
            log.info("Running %s on %s", self.opts["start_action"], vm_["name"])
            with salt.client.get_local_client(mopts=self.opts) as client:
                action_out = client.cmd(
                    vm_["name"],
                    self.opts["start_action"],
                    timeout=self.opts["timeout"] * 60,
                )
            output["ret"] = action_out
        return output

    @staticmethod
    def vm_config(name, main, provider, profile, overrides):
        """
        Create vm config.

        :param str name: The name of the vm
        :param dict main: The main cloud config
        :param dict provider: The provider config
        :param dict profile: The profile config
        :param dict overrides: a special dict that carries per-node options overrides (see CloudConfig:profile() documentation)
        """
        vm = main.copy()
        vm = salt.utils.dictupdate.update(vm, provider)
        vm = salt.utils.dictupdate.update(vm, profile)
        vm = salt.utils.dictupdate.update(vm, overrides.get(name, {}))
        vm["name"] = name
        return vm

    def extras(self, extra_):
        """
        Extra actions
        """
        output = {}

        alias, driver = extra_["provider"].split(":")
        fun = "{}.{}".format(driver, extra_["action"])
        if fun not in self.clouds:
            log.error(
                "Creating '%s' using '%s' as the provider "
                "cannot complete since '%s' is not available",
                extra_["name"],
                extra_["provider"],
                driver,
            )
            return

        try:
            with salt.utils.context.func_globals_inject(
                self.clouds[fun], __active_provider_name__=extra_["provider"]
            ):
                output = self.clouds[fun](**extra_)
        except KeyError as exc:
            log.exception(
                "Failed to perform %s.%s on %s. Configuration value %s needs to be set",
                extra_["provider"],
                extra_["action"],
                extra_["name"],
                exc,
            )
        return output

    def run_profile(self, profile, names, vm_overrides=None):
        """
        Parse over the options passed on the command line and determine how to
        handle them
        """
        if profile not in self.opts["profiles"]:
            msg = f"Profile {profile} is not defined"
            log.error(msg)
            return {"Error": msg}

        ret = {}
        if not vm_overrides:
            vm_overrides = {}

        try:
            with salt.utils.files.fopen(self.opts["conf_file"], "r") as mcc:
                main_cloud_config = salt.utils.yaml.safe_load(mcc)
            if not main_cloud_config:
                main_cloud_config = {}
        except KeyError:
            main_cloud_config = {}
        except OSError:
            main_cloud_config = {}

        if main_cloud_config is None:
            main_cloud_config = {}

        mapped_providers = self.map_providers_parallel()
        profile_details = self.opts["profiles"][profile]
        vms = {}
        for prov, val in mapped_providers.items():
            prov_name = next(iter(val))
            for node in mapped_providers[prov][prov_name]:
                vms[node] = mapped_providers[prov][prov_name][node]
                vms[node]["provider"] = prov
                vms[node]["driver"] = prov_name
        alias, driver = profile_details["provider"].split(":")

        provider_details = self.opts["providers"][alias][driver].copy()
        del provider_details["profiles"]

        for name in names:
            if name in vms:
                prov = vms[name]["provider"]
                driv = vms[name]["driver"]
                msg = f"{name} already exists under {prov}:{driv}"
                log.error(msg)
                ret[name] = {"Error": msg}
                continue

            vm_ = self.vm_config(
                name,
                main_cloud_config,
                provider_details,
                profile_details,
                vm_overrides,
            )
            if self.opts["parallel"]:
                process = multiprocessing.Process(target=self.create, args=(vm_,))
                process.start()
                ret[name] = {
                    "Provisioning": "VM being provisioned in parallel. PID: {}".format(
                        process.pid
                    )
                }
                continue

            try:
                # No need to inject __active_provider_name__ into the context
                # here because self.create takes care of that
                ret[name] = self.create(vm_)
                if not ret[name]:
                    ret[name] = {"Error": "Failed to deploy VM"}
                    if len(names) == 1:
                        raise SaltCloudSystemExit("Failed to deploy VM")
                    continue
                if self.opts.get("show_deploy_args", False) is False:
                    if isinstance(ret[name], dict):
                        ret[name].pop("deploy_kwargs", None)
            except (SaltCloudSystemExit, SaltCloudConfigError) as exc:
                if len(names) == 1:
                    raise
                ret[name] = {"Error": str(exc)}

        return ret

    def do_action(self, names, kwargs):
        """
        Perform an action on a VM which may be specific to this cloud provider
        """
        ret = {}
        invalid_functions = {}
        names = set(names)

        for alias, drivers in self.map_providers_parallel().items():
            if not names:
                break
            for driver, vms in drivers.items():
                if not names:
                    break
                valid_function = True
                fun = "{}.{}".format(driver, self.opts["action"])
                if fun not in self.clouds:
                    log.info("'%s()' is not available. Not actioning...", fun)
                    valid_function = False
                for vm_name, vm_details in vms.items():
                    if not names:
                        break
                    if vm_name not in names:
                        if not isinstance(vm_details, dict):
                            vm_details = {}
                        if "id" in vm_details and vm_details["id"] in names:
                            vm_name = vm_details["id"]
                        else:
                            log.debug(
                                "vm:%s in provider:%s is not in name list:'%s'",
                                vm_name,
                                driver,
                                names,
                            )
                            continue

                    # Build the dictionary of invalid functions with their associated VMs.
                    if valid_function is False:
                        if invalid_functions.get(fun) is None:
                            invalid_functions.update({fun: []})
                        invalid_functions[fun].append(vm_name)
                        continue

                    with salt.utils.context.func_globals_inject(
                        self.clouds[fun],
                        __active_provider_name__=":".join([alias, driver]),
                    ):
                        if alias not in ret:
                            ret[alias] = {}
                        if driver not in ret[alias]:
                            ret[alias][driver] = {}

                        # Clean kwargs of "__pub_*" data before running the cloud action call.
                        # Prevents calling positional "kwarg" arg before "call" when no kwarg
                        # argument is present in the cloud driver function's arg spec.
                        kwargs = salt.utils.args.clean_kwargs(**kwargs)

                        if kwargs:
                            ret[alias][driver][vm_name] = self.clouds[fun](
                                vm_name, kwargs, call="action"
                            )
                        else:
                            ret[alias][driver][vm_name] = self.clouds[fun](
                                vm_name, call="action"
                            )
                        names.remove(vm_name)

        # Set the return information for the VMs listed in the invalid_functions dict.
        missing_vms = set()
        if invalid_functions:
            ret["Invalid Actions"] = invalid_functions
            invalid_func_vms = set()
            for key, val in invalid_functions.items():
                invalid_func_vms = invalid_func_vms.union(set(val))

            # Find the VMs that are in names, but not in set of invalid functions.
            missing_vms = names.difference(invalid_func_vms)
            if missing_vms:
                ret["Not Found"] = list(missing_vms)
                ret["Not Actioned/Not Running"] = list(names)

        if not names:
            return ret

        # Don't return missing VM information for invalid functions until after we've had a
        # Chance to return successful actions. If a function is valid for one driver, but
        # Not another, we want to make sure the successful action is returned properly.
        if missing_vms:
            return ret

        # If we reach this point, the Not Actioned and Not Found lists will be the same,
        # But we want to list both for clarity/consistency with the invalid functions lists.
        ret["Not Actioned/Not Running"] = list(names)
        ret["Not Found"] = list(names)
        return ret

    def do_function(self, prov, func, kwargs):
        """
        Perform a function against a cloud provider
        """
        matches = self.lookup_providers(prov)
        if len(matches) > 1:
            raise SaltCloudSystemExit(
                "More than one results matched '{}'. Please specify one of: {}".format(
                    prov,
                    ", ".join([f"{alias}:{driver}" for (alias, driver) in matches]),
                )
            )

        alias, driver = matches.pop()
        fun = f"{driver}.{func}"
        if fun not in self.clouds:
            raise SaltCloudSystemExit(
                "The '{}' cloud provider alias, for the '{}' driver, does "
                "not define the function '{}'".format(alias, driver, func)
            )

        log.debug("Trying to execute '%s' with the following kwargs: %s", fun, kwargs)

        with salt.utils.context.func_globals_inject(
            self.clouds[fun], __active_provider_name__=":".join([alias, driver])
        ):
            if kwargs:
                return {
                    alias: {driver: self.clouds[fun](call="function", kwargs=kwargs)}
                }
            return {alias: {driver: self.clouds[fun](call="function")}}

    def __filter_non_working_providers(self):
        """
        Remove any mis-configured cloud providers from the available listing
        """
        for alias, drivers in self.opts["providers"].copy().items():
            for driver in drivers.copy():
                fun = f"{driver}.get_configured_provider"
                if fun not in self.clouds:
                    # Mis-configured provider that got removed?
                    log.warning(
                        "The cloud driver, '%s', configured under the "
                        "'%s' cloud provider alias, could not be loaded. "
                        "Please check your provider configuration files and "
                        "ensure all required dependencies are installed "
                        "for the '%s' driver.\n"
                        "In rare cases, this could indicate the '%s()' "
                        "function could not be found.\nRemoving '%s' from "
                        "the available providers list",
                        driver,
                        alias,
                        driver,
                        fun,
                        driver,
                    )
                    self.opts["providers"][alias].pop(driver)

                    if alias not in self.opts["providers"]:
                        continue

                    if not self.opts["providers"][alias]:
                        self.opts["providers"].pop(alias)
                    continue

                with salt.utils.context.func_globals_inject(
                    self.clouds[fun], __active_provider_name__=":".join([alias, driver])
                ):
                    if self.clouds[fun]() is False:
                        log.warning(
                            "The cloud driver, '%s', configured under the "
                            "'%s' cloud provider alias is not properly "
                            "configured. Removing it from the available "
                            "providers list.",
                            driver,
                            alias,
                        )
                        self.opts["providers"][alias].pop(driver)

            if alias not in self.opts["providers"]:
                continue

            if not self.opts["providers"][alias]:
                self.opts["providers"].pop(alias)


class Map(Cloud):
    """
    Create a VM stateful map execution object
    """

    def __init__(self, opts):
        Cloud.__init__(self, opts)
        self.rendered_map = self.read()

    def interpolated_map(self, query="list_nodes", cached=False):
        rendered_map = self.read().copy()
        interpolated_map = {}

        for profile, mapped_vms in rendered_map.items():
            names = set(mapped_vms)
            if profile not in self.opts["profiles"]:
                if "Errors" not in interpolated_map:
                    interpolated_map["Errors"] = {}
                msg = (
                    "No provider for the mapped '{}' profile was found. "
                    "Skipped VMS: {}".format(profile, ", ".join(names))
                )
                log.info(msg)
                interpolated_map["Errors"][profile] = msg
                continue

            matching = self.get_running_by_names(names, query, cached)
            for alias, drivers in matching.items():
                for driver, vms in drivers.items():
                    for vm_name, vm_details in vms.items():
                        if alias not in interpolated_map:
                            interpolated_map[alias] = {}
                        if driver not in interpolated_map[alias]:
                            interpolated_map[alias][driver] = {}
                        interpolated_map[alias][driver][vm_name] = vm_details
                        try:
                            names.remove(vm_name)
                        except KeyError:
                            # If it's not there, then our job is already done
                            pass

            if not names:
                continue

            profile_details = self.opts["profiles"][profile]
            alias, driver = profile_details["provider"].split(":")
            for vm_name in names:
                if alias not in interpolated_map:
                    interpolated_map[alias] = {}
                if driver not in interpolated_map[alias]:
                    interpolated_map[alias][driver] = {}
                interpolated_map[alias][driver][vm_name] = "Absent"

        return interpolated_map

    def delete_map(self, query=None):
        query_map = self.interpolated_map(query=query)
        for alias, drivers in query_map.copy().items():
            if alias == "Errors":
                continue
            for driver, vms in drivers.copy().items():
                for vm_name, vm_details in vms.copy().items():
                    if vm_details == "Absent":
                        query_map[alias][driver].pop(vm_name)
                if not query_map[alias][driver]:
                    query_map[alias].pop(driver)
            if not query_map[alias]:
                query_map.pop(alias)
        return query_map

    def get_vmnames_by_action(self, action):
        query_map = self.interpolated_map("list_nodes")
        matching_states = {
            "start": ["stopped"],
            "stop": ["running", "active"],
            "reboot": ["running", "active"],
        }
        vm_names = []
        for alias, drivers in query_map.items():
            for driver, vms in drivers.items():
                for vm_name, vm_details in vms.items():
                    # Only certain actions are support in to use in this case. Those actions are the
                    # "Global" salt-cloud actions defined in the "matching_states" dictionary above.
                    # If a more specific action is passed in, we shouldn't stack-trace - exit gracefully.
                    try:
                        state_action = matching_states[action]
                    except KeyError:
                        log.error(
                            "The use of '%s' as an action is not supported "
                            "in this context. Only 'start', 'stop', and "
                            "'reboot' are supported options.",
                            action,
                        )
                        raise SaltCloudException()
                    if (
                        vm_details != "Absent"
                        and vm_details["state"].lower() in state_action
                    ):
                        vm_names.append(vm_name)
        return vm_names

    def read(self):
        """
        Read in the specified map and return the map structure
        """
        map_ = None
        if self.opts.get("map", None) is None:
            if self.opts.get("map_data", None) is None:
                if self.opts.get("map_pillar", None) is None:
                    pass
                elif self.opts.get("map_pillar") not in self.opts.get("maps"):
                    log.error(
                        "The specified map not found in pillar at 'cloud:maps:%s'",
                        self.opts["map_pillar"],
                    )
                    raise SaltCloudNotFound()
                else:
                    # 'map_pillar' is provided, try to use it
                    map_ = self.opts["maps"][self.opts.get("map_pillar")]
            else:
                # 'map_data' is provided, try to use it
                map_ = self.opts["map_data"]
        else:
            # 'map' is provided, try to use it
            local_minion_opts = copy.deepcopy(self.opts)
            local_minion_opts["file_client"] = "local"
            self.minion = salt.minion.MasterMinion(local_minion_opts)

            if not os.path.isfile(self.opts["map"]):
                if not (self.opts["map"]).startswith("salt://"):
                    log.error(
                        "The specified map file does not exist: '%s'", self.opts["map"]
                    )
                    raise SaltCloudNotFound()
            if (self.opts["map"]).startswith("salt://"):
                cached_map = self.minion.functions["cp.cache_file"](self.opts["map"])
            else:
                cached_map = self.opts["map"]
            try:
                renderer = self.opts.get("renderer", "jinja|yaml")
                rend = salt.loader.render(self.opts, {})
                blacklist = self.opts.get("renderer_blacklist")
                whitelist = self.opts.get("renderer_whitelist")
                map_ = compile_template(
                    cached_map, rend, renderer, blacklist, whitelist
                )
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Rendering map %s failed, render error:\n%s",
                    self.opts["map"],
                    exc,
                    exc_info_on_loglevel=logging.DEBUG,
                )
                return {}

            if "include" in map_:
                map_ = salt.config.include_config(map_, self.opts["map"], verbose=False)

        if not map_:
            return {}

        # Create expected data format if needed
        for profile, mapped in map_.copy().items():
            if isinstance(mapped, (list, tuple)):
                entries = {}
                for mapping in mapped:
                    if isinstance(mapping, str):
                        # Foo:
                        #   - bar1
                        #   - bar2
                        mapping = {mapping: None}
                    for name, overrides in mapping.items():
                        if overrides is None or isinstance(overrides, bool):
                            # Foo:
                            #   - bar1:
                            #   - bar2:
                            overrides = {}
                        try:
                            overrides.setdefault("name", name)
                        except AttributeError:
                            log.error(
                                "Cannot use 'name' as a minion id in a cloud map as it"
                                " is a reserved word. Please change 'name' to a"
                                " different minion id reference."
                            )
                            return {}
                        entries[name] = overrides
                map_[profile] = entries
                continue

            if isinstance(mapped, dict):
                # Convert the dictionary mapping to a list of dictionaries
                # Foo:
                #  bar1:
                #    grains:
                #      foo: bar
                #  bar2:
                #    grains:
                #      foo: bar
                entries = {}
                for name, overrides in mapped.items():
                    overrides.setdefault("name", name)
                    entries[name] = overrides
                map_[profile] = entries
                continue

            if isinstance(mapped, str):
                # If it's a single string entry, let's make iterable because of
                # the next step
                mapped = [mapped]

            map_[profile] = {}
            for name in mapped:
                map_[profile][name] = {"name": name}
        return map_

    def _has_loop(self, dmap, seen=None, val=None):
        if seen is None:
            for values in dmap["create"].values():
                seen = []
                try:
                    machines = values["requires"]
                except KeyError:
                    machines = []
                for machine in machines:
                    if self._has_loop(dmap, seen=list(seen), val=machine):
                        return True
        else:
            if val in seen:
                return True

            seen.append(val)
            try:
                machines = dmap["create"][val]["requires"]
            except KeyError:
                machines = []

            for machine in machines:
                if self._has_loop(dmap, seen=list(seen), val=machine):
                    return True
        return False

    def _calcdep(self, dmap, machine, data, level):
        try:
            deplist = data["requires"]
        except KeyError:
            return level
        levels = []
        for name in deplist:
            try:
                data = dmap["create"][name]
            except KeyError:
                try:
                    data = dmap["existing"][name]
                except KeyError:
                    msg = "Missing dependency in cloud map"
                    log.error(msg)
                    raise SaltCloudException(msg)
            levels.append(self._calcdep(dmap, name, data, level))
        level = max(levels) + 1
        return level

    def map_data(self, cached=False):
        """
        Create a data map of what to execute on
        """
        ret = {"create": {}}
        pmap = self.map_providers_parallel(cached=cached)
        exist = set()
        defined = set()
        rendered_map = copy.deepcopy(self.rendered_map)
        for profile_name, nodes in rendered_map.items():
            if profile_name not in self.opts["profiles"]:
                msg = (
                    "The required profile, '{}', defined in the map "
                    "does not exist. The defined nodes, {}, will not "
                    "be created.".format(
                        profile_name, ", ".join(f"'{node}'" for node in nodes)
                    )
                )
                log.error(msg)
                if "errors" not in ret:
                    ret["errors"] = {}
                ret["errors"][profile_name] = msg
                continue

            profile_data = self.opts["profiles"].get(profile_name)

            for nodename, overrides in nodes.items():
                # Get associated provider data, in case something like size
                # or image is specified in the provider file. See issue #32510.
                if (
                    "provider" in overrides
                    and overrides["provider"] != profile_data["provider"]
                ):
                    alias, driver = overrides.get("provider").split(":")
                else:
                    alias, driver = profile_data.get("provider").split(":")

                provider_details = copy.deepcopy(self.opts["providers"][alias][driver])
                del provider_details["profiles"]

                # Update the provider details information with profile data
                # Profile data and node overrides should override provider data, if defined.
                # This keeps map file data definitions consistent with -p usage.
                salt.utils.dictupdate.update(provider_details, profile_data)
                nodedata = copy.deepcopy(provider_details)

                # Update profile data with the map overrides
                for setting in ("grains", "master", "minion", "volumes", "requires"):
                    deprecated = f"map_{setting}"
                    if deprecated in overrides:
                        log.warning(
                            "The use of '%s' on the '%s' mapping has "
                            "been deprecated. The preferred way now is to "
                            "just define '%s'. For now, salt-cloud will do "
                            "the proper thing and convert the deprecated "
                            "mapping into the preferred one.",
                            deprecated,
                            nodename,
                            setting,
                        )
                        overrides[setting] = overrides.pop(deprecated)

                # merge minion grains from map file
                if (
                    "minion" in overrides
                    and "minion" in nodedata
                    and "grains" in overrides["minion"]
                    and "grains" in nodedata["minion"]
                ):
                    nodedata["minion"]["grains"].update(overrides["minion"]["grains"])
                    del overrides["minion"]["grains"]
                    # remove minion key if now is empty dict
                    if not overrides["minion"]:
                        del overrides["minion"]

                nodedata = salt.utils.dictupdate.update(nodedata, overrides)
                # Add the computed information to the return data
                ret["create"][nodename] = nodedata
                # Add the node name to the defined set
                alias, driver = nodedata["provider"].split(":")
                defined.add((alias, driver, nodename))

        def get_matching_by_name(name):
            matches = {}
            for alias, drivers in pmap.items():
                for driver, vms in drivers.items():
                    for vm_name, details in vms.items():
                        if vm_name == name and driver not in matches:
                            matches[driver] = details["state"]
            return matches

        for alias, drivers in pmap.items():
            for driver, vms in drivers.items():
                for name, details in vms.items():
                    exist.add((alias, driver, name))
                    if name not in ret["create"]:
                        continue

                    # The machine is set to be created. Does it already exist?
                    matching = get_matching_by_name(name)
                    if not matching:
                        continue

                    # A machine by the same name exists
                    for item in matching:
                        if name not in ret["create"]:
                            # Machine already removed
                            break

                        log.warning(
                            "%r already exists, removing from the create map.", name
                        )

                        if "existing" not in ret:
                            ret["existing"] = {}
                        ret["existing"][name] = ret["create"].pop(name)

        if "hard" in self.opts and self.opts["hard"]:
            if self.opts["enable_hard_maps"] is False:
                raise SaltCloudSystemExit(
                    "The --hard map can be extremely dangerous to use, "
                    "and therefore must explicitly be enabled in the main "
                    "configuration file, by setting 'enable_hard_maps' "
                    "to True"
                )

            # Hard maps are enabled, Look for the items to delete.
            ret["destroy"] = exist.difference(defined)
        return ret

    def run_map(self, dmap):
        """
        Execute the contents of the VM map
        """
        if self._has_loop(dmap):
            msg = "Uh-oh, that cloud map has a dependency loop!"
            log.error(msg)
            raise SaltCloudException(msg)
        # Go through the create list and calc dependencies
        for key, val in dmap["create"].items():
            log.info("Calculating dependencies for %s", key)
            level = 0
            level = self._calcdep(dmap, key, val, level)
            log.debug("Got execution order %s for %s", level, key)
            dmap["create"][key]["level"] = level

        try:
            existing_list = dmap["existing"].items()
        except KeyError:
            existing_list = {}.items()

        for key, val in existing_list:
            log.info("Calculating dependencies for %s", key)
            level = 0
            level = self._calcdep(dmap, key, val, level)
            log.debug("Got execution order %s for %s", level, key)
            dmap["existing"][key]["level"] = level

        # Now sort the create list based on dependencies
        create_list = sorted(dmap["create"].items(), key=lambda x: x[1]["level"])
        output = {}
        if self.opts["parallel"]:
            parallel_data = []
        master_name = None
        master_minion_name = None
        master_host = None
        master_finger = None
        try:
            master_name, master_profile = next(
                (
                    (name, profile)
                    for name, profile in create_list
                    if profile.get("make_master", False) is True
                )
            )
            master_minion_name = master_name
            log.debug("Creating new master '%s'", master_name)
            if (
                salt.config.get_cloud_config_value("deploy", master_profile, self.opts)
                is False
            ):
                raise SaltCloudSystemExit(
                    "Cannot proceed with 'make_master' when salt deployment "
                    "is disabled(ex: --no-deploy)."
                )

            # Generate the master keys
            log.debug("Generating master keys for '%s'", master_profile["name"])
            priv, pub = salt.utils.cloud.gen_keys(
                salt.config.get_cloud_config_value("keysize", master_profile, self.opts)
            )
            master_profile["master_pub"] = pub
            master_profile["master_pem"] = priv

            # Generate the fingerprint of the master pubkey in order to
            # mitigate man-in-the-middle attacks
            master_temp_pub = salt.utils.files.mkstemp()
            with salt.utils.files.fopen(master_temp_pub, "w") as mtp:
                mtp.write(pub)
            master_finger = salt.utils.crypt.pem_finger(
                master_temp_pub, sum_type=self.opts["hash_type"]
            )
            os.unlink(master_temp_pub)

            if master_profile.get("make_minion", True) is True:
                master_profile.setdefault("minion", {})
                if "id" in master_profile["minion"]:
                    master_minion_name = master_profile["minion"]["id"]
                # Set this minion's master as local if the user has not set it
                if "master" not in master_profile["minion"]:
                    master_profile["minion"]["master"] = "127.0.0.1"
                    if master_finger is not None:
                        master_profile["master_finger"] = master_finger

            # Generate the minion keys to pre-seed the master:
            for name, profile in create_list:
                make_minion = salt.config.get_cloud_config_value(
                    "make_minion", profile, self.opts, default=True
                )
                if make_minion is False:
                    continue

                log.debug("Generating minion keys for '%s'", profile["name"])
                priv, pub = salt.utils.cloud.gen_keys(
                    salt.config.get_cloud_config_value("keysize", profile, self.opts)
                )
                profile["pub_key"] = pub
                profile["priv_key"] = priv
                # Store the minion's public key in order to be pre-seeded in
                # the master
                master_profile.setdefault("preseed_minion_keys", {})
                master_profile["preseed_minion_keys"].update({name: pub})

            local_master = False
            if (
                master_profile["minion"].get("local_master", False)
                and master_profile["minion"].get("master", None) is not None
            ):
                # The minion is explicitly defining a master and it's
                # explicitly saying it's the local one
                local_master = True

            out = self.create(master_profile, local_master=local_master)

            if not isinstance(out, dict):
                log.debug("Master creation details is not a dictionary: %s", out)

            elif "Errors" in out:
                raise SaltCloudSystemExit(
                    "An error occurred while creating the master, not "
                    "continuing: {}".format(out["Errors"])
                )

            deploy_kwargs = (
                self.opts.get("show_deploy_args", False) is True
                and
                # Get the needed data
                out.get("deploy_kwargs", {})
                or
                # Strip the deploy_kwargs from the returned data since we don't
                # want it shown in the console.
                out.pop("deploy_kwargs", {})
            )

            master_host = deploy_kwargs.get(
                "salt_host", deploy_kwargs.get("host", None)
            )
            if master_host is None:
                raise SaltCloudSystemExit(
                    "Host for new master {} was not found, aborting map".format(
                        master_name
                    )
                )
            output[master_name] = out
        except StopIteration:
            log.debug("No make_master found in map")
            # Local master?
            # Generate the fingerprint of the master pubkey in order to
            # mitigate man-in-the-middle attacks
            master_pub = os.path.join(self.opts["pki_dir"], "master.pub")
            if os.path.isfile(master_pub):
                master_finger = salt.utils.crypt.pem_finger(
                    master_pub, sum_type=self.opts["hash_type"]
                )

        opts = self.opts.copy()
        if self.opts["parallel"]:
            # Force display_ssh_output to be False since the console will
            # need to be reset afterwards
            log.info(
                "Since parallel deployment is in use, ssh console output "
                "is disabled. All ssh output will be logged though"
            )
            opts["display_ssh_output"] = False

        local_master = master_name is None

        for name, profile in create_list:
            if name in (master_name, master_minion_name):
                # Already deployed, it's the master's minion
                continue

            if (
                "minion" in profile
                and profile["minion"].get("local_master", False)
                and profile["minion"].get("master", None) is not None
            ):
                # The minion is explicitly defining a master and it's
                # explicitly saying it's the local one
                local_master = True

            if master_finger is not None and local_master is False:
                profile["master_finger"] = master_finger

            if master_host is not None:
                profile.setdefault("minion", {})
                profile["minion"].setdefault("master", master_host)

            if self.opts["parallel"]:
                parallel_data.append(
                    {
                        "opts": opts,
                        "name": name,
                        "profile": profile,
                        "local_master": local_master,
                    }
                )
                continue

            # Not deploying in parallel
            try:
                output[name] = self.create(profile, local_master=local_master)
                if (
                    self.opts.get("show_deploy_args", False) is False
                    and "deploy_kwargs" in output
                    and isinstance(output[name], dict)
                ):
                    output[name].pop("deploy_kwargs", None)
            except SaltCloudException as exc:
                log.error(
                    "Failed to deploy '%s'. Error: %s",
                    name,
                    exc,
                    exc_info_on_loglevel=logging.DEBUG,
                )
                output[name] = {"Error": str(exc)}

        for name in dmap.get("destroy", ()):
            output[name] = self.destroy(name)

        if self.opts["parallel"] and parallel_data:
            if "pool_size" in self.opts:
                pool_size = self.opts["pool_size"]
            else:
                pool_size = len(parallel_data)
            log.info("Cloud pool size: %s", pool_size)
            output_multip = enter_mainloop(
                _create_multiprocessing, parallel_data, pool_size=pool_size
            )
            # We have deployed in parallel, now do start action in
            # correct order based on dependencies.
            if self.opts["start_action"]:
                actionlist = []
                grp = -1
                for key, val in groupby(dmap["create"].values(), lambda x: x["level"]):
                    actionlist.append([])
                    grp += 1
                    for item in val:
                        actionlist[grp].append(item["name"])

                out = {}
                for group in actionlist:
                    log.info(
                        "Running %s on %s", self.opts["start_action"], ", ".join(group)
                    )
                    with salt.client.get_local_client() as client:
                        out.update(
                            client.cmd(
                                ",".join(group),
                                self.opts["start_action"],
                                timeout=self.opts["timeout"] * 60,
                                tgt_type="list",
                            )
                        )
                for obj in output_multip:
                    next(iter(obj.values()))["ret"] = out[next(iter(obj.keys()))]
                    output.update(obj)
            else:
                for obj in output_multip:
                    output.update(obj)

        return output


def init_pool_worker():
    """
    Make every worker ignore KeyboarInterrup's since it will be handled by the
    parent process.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def create_multiprocessing(parallel_data, queue=None):
    """
    This function will be called from another process when running a map in
    parallel mode. The result from the create is always a json object.
    """
    parallel_data["opts"]["output"] = "json"
    cloud = Cloud(parallel_data["opts"])
    try:
        output = cloud.create(
            parallel_data["profile"], local_master=parallel_data["local_master"]
        )
    except SaltCloudException as exc:
        log.error(
            "Failed to deploy '%s'. Error: %s",
            parallel_data["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return {parallel_data["name"]: {"Error": str(exc)}}

    if parallel_data["opts"].get("show_deploy_args", False) is False and isinstance(
        output, dict
    ):
        output.pop("deploy_kwargs", None)

    return {parallel_data["name"]: salt.utils.data.simple_types_filter(output)}


def destroy_multiprocessing(parallel_data, queue=None):
    """
    This function will be called from another process when running a map in
    parallel mode. The result from the destroy is always a json object.
    """
    parallel_data["opts"]["output"] = "json"
    clouds = salt.loader.clouds(parallel_data["opts"])

    try:
        fun = clouds["{}.destroy".format(parallel_data["driver"])]
        with salt.utils.context.func_globals_inject(
            fun,
            __active_provider_name__=":".join(
                [parallel_data["alias"], parallel_data["driver"]]
            ),
        ):
            output = fun(parallel_data["name"])

    except SaltCloudException as exc:
        log.error(
            "Failed to destroy %s. Error: %s",
            parallel_data["name"],
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return {parallel_data["name"]: {"Error": str(exc)}}

    return {parallel_data["name"]: salt.utils.data.simple_types_filter(output)}


def run_parallel_map_providers_query(data, queue=None):
    """
    This function will be called from another process when building the
    providers map.
    """
    cloud = Cloud(data["opts"])
    try:
        with salt.utils.context.func_globals_inject(
            cloud.clouds[data["fun"]],
            __active_provider_name__=":".join([data["alias"], data["driver"]]),
        ):
            return (
                data["alias"],
                data["driver"],
                salt.utils.data.simple_types_filter(cloud.clouds[data["fun"]]()),
            )
    except Exception as err:  # pylint: disable=broad-except
        log.debug(
            "Failed to execute '%s()' while querying for running nodes: %s",
            data["fun"],
            err,
            exc_info_on_loglevel=logging.DEBUG,
        )
        # Failed to communicate with the provider, don't list any nodes
        return data["alias"], data["driver"], ()


# for pickle and multiprocessing, we can't use directly decorators
def _run_parallel_map_providers_query(*args, **kw):
    return communicator(run_parallel_map_providers_query)(*args[0], **kw)


def _destroy_multiprocessing(*args, **kw):
    return communicator(destroy_multiprocessing)(*args[0], **kw)


def _create_multiprocessing(*args, **kw):
    return communicator(create_multiprocessing)(*args[0], **kw)



"""Utilities to manage open connections."""

import io
import os
import socket
import threading
import time
import selectors
from contextlib import suppress

from . import errors
from ._compat import IS_WINDOWS
from .makefile import MakeFile

try:
    import fcntl
except ImportError:
    try:
        from ctypes import windll, WinError
        import ctypes.wintypes
        _SetHandleInformation = windll.kernel32.SetHandleInformation
        _SetHandleInformation.argtypes = [
            ctypes.wintypes.HANDLE,
            ctypes.wintypes.DWORD,
            ctypes.wintypes.DWORD,
        ]
        _SetHandleInformation.restype = ctypes.wintypes.BOOL
    except ImportError:
        def prevent_socket_inheritance(sock):
            """Stub inheritance prevention.

            Dummy function, since neither fcntl nor ctypes are available.
            """
            pass
    else:
        def prevent_socket_inheritance(sock):
            """Mark the given socket fd as non-inheritable (Windows)."""
            if not _SetHandleInformation(sock.fileno(), 1, 0):
                raise WinError()
else:
    def prevent_socket_inheritance(sock):
        """Mark the given socket fd as non-inheritable (POSIX)."""
        fd = sock.fileno()
        old_flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        fcntl.fcntl(fd, fcntl.F_SETFD, old_flags | fcntl.FD_CLOEXEC)


class _ThreadsafeSelector:
    """Thread-safe wrapper around a DefaultSelector.

    There are 2 thread contexts in which it may be accessed:
      * the selector thread
      * one of the worker threads in workers/threadpool.py

    The expected read/write patterns are:
      * :py:func:`~iter`: selector thread
      * :py:meth:`register`: selector thread and threadpool,
        via :py:meth:`~cheroot.workers.threadpool.ThreadPool.put`
      * :py:meth:`unregister`: selector thread only

    Notably, this means :py:class:`_ThreadsafeSelector` never needs to worry
    that connections will be removed behind its back.

    The lock is held when iterating or modifying the selector but is not
    required when :py:meth:`select()ing <selectors.BaseSelector.select>` on it.
    """

    def __init__(self):
        self._selector = selectors.DefaultSelector()
        self._lock = threading.Lock()

    def __len__(self):
        with self._lock:
            return len(self._selector.get_map() or {})

    @property
    def connections(self):
        """Retrieve connections registered with the selector."""
        with self._lock:
            mapping = self._selector.get_map() or {}
            for _, (_, sock_fd, _, conn) in mapping.items():
                yield (sock_fd, conn)

    def register(self, fileobj, events, data=None):
        """Register ``fileobj`` with the selector."""
        with self._lock:
            return self._selector.register(fileobj, events, data)

    def unregister(self, fileobj):
        """Unregister ``fileobj`` from the selector."""
        with self._lock:
            return self._selector.unregister(fileobj)

    def select(self, timeout=None):
        """Return socket fd and data pairs from selectors.select call.

        Returns entries ready to read in the form:
            (socket_file_descriptor, connection)
        """
        return (
            (key.fd, key.data)
            for key, _ in self._selector.select(timeout=timeout)
        )

    def close(self):
        """Close the selector."""
        with self._lock:
            self._selector.close()


class ConnectionManager:
    """Class which manages HTTPConnection objects.

    This is for connections which are being kept-alive for follow-up requests.
    """

    def __init__(self, server):
        """Initialize ConnectionManager object.

        Args:
            server (cheroot.server.HTTPServer): web server object
                that uses this ConnectionManager instance.
        """
        self._serving = False
        self._stop_requested = False

        self.server = server
        self._selector = _ThreadsafeSelector()

        self._selector.register(
            server.socket.fileno(),
            selectors.EVENT_READ, data=server,
        )

    def put(self, conn):
        """Put idle connection into the ConnectionManager to be managed.

        :param conn: HTTP connection to be managed
        :type conn: cheroot.server.HTTPConnection
        """
        conn.last_used = time.time()
        # if this conn doesn't have any more data waiting to be read,
        # register it with the selector.
        if conn.rfile.has_data():
            self.server.process_conn(conn)
        else:
            self._selector.register(
                conn.socket.fileno(), selectors.EVENT_READ, data=conn,
            )

    def _expire(self, threshold):
        r"""Expire least recently used connections.

        :param threshold: Connections that have not been used within this \
                          duration (in seconds), are considered expired and \
                          are closed and removed.
        :type threshold: float

        This should be called periodically.
        """
        # find any connections still registered with the selector
        # that have not been active recently enough.
        timed_out_connections = [
            (sock_fd, conn)
            for (sock_fd, conn) in self._selector.connections
            if conn != self.server and conn.last_used < threshold
        ]
        for sock_fd, conn in timed_out_connections:
            self._selector.unregister(sock_fd)
            conn.close()

    def stop(self):
        """Stop the selector loop in run() synchronously.

        May take up to half a second.
        """
        self._stop_requested = True
        while self._serving:
            time.sleep(0.01)

    def run(self, expiration_interval):
        """Run the connections selector indefinitely.

        Args:
            expiration_interval (float): Interval, in seconds, at which
                connections will be checked for expiration.

        Connections that are ready to process are submitted via
        self.server.process_conn()

        Connections submitted for processing must be `put()`
        back if they should be examined again for another request.

        Can be shut down by calling `stop()`.
        """
        self._serving = True
        try:
            self._run(expiration_interval)
        finally:
            self._serving = False

    def _run(self, expiration_interval):
        r"""Run connection handler loop until stop was requested.

        :param expiration_interval: Interval, in seconds, at which \
                                    connections will be checked for \
                                    expiration.
        :type expiration_interval: float

        Use ``expiration_interval`` as ``select()`` timeout
        to assure expired connections are closed in time.

        On Windows cap the timeout to 0.05 seconds
        as ``select()`` does not return when a socket is ready.
        """
        last_expiration_check = time.time()
        if IS_WINDOWS:
            # 0.05 seconds are used as an empirically obtained balance between
            # max connection delay and idle system load. Benchmarks show a
            # mean processing time per connection of ~0.03 seconds on Linux
            # and with 0.01 seconds timeout on Windows:
            # https://github.com/cherrypy/cheroot/pull/352
            # While this highly depends on system and hardware, 0.05 seconds
            # max delay should hence usually not significantly increase the
            # mean time/delay per connection, but significantly reduce idle
            # system load by reducing socket loops to 1/5 with 0.01 seconds.
            select_timeout = min(expiration_interval, 0.05)
        else:
            select_timeout = expiration_interval

        while not self._stop_requested:
            try:
                active_list = self._selector.select(timeout=select_timeout)
            except OSError:
                self._remove_invalid_sockets()
                continue

            for (sock_fd, conn) in active_list:
                if conn is self.server:
                    # New connection
                    new_conn = self._from_server_socket(self.server.socket)
                    if new_conn is not None:
                        self.server.process_conn(new_conn)
                else:
                    # unregister connection from the selector until the server
                    # has read from it and returned it via put()
                    self._selector.unregister(sock_fd)
                    self.server.process_conn(conn)

            now = time.time()
            if (now - last_expiration_check) > expiration_interval:
                self._expire(threshold=now - self.server.timeout)
                last_expiration_check = now

    def _remove_invalid_sockets(self):
        """Clean up the resources of any broken connections.

        This method attempts to detect any connections in an invalid state,
        unregisters them from the selector and closes the file descriptors of
        the corresponding network sockets where possible.
        """
        invalid_conns = []
        for sock_fd, conn in self._selector.connections:
            if conn is self.server:
                continue

            try:
                os.fstat(sock_fd)
            except OSError:
                invalid_conns.append((sock_fd, conn))

        for sock_fd, conn in invalid_conns:
            self._selector.unregister(sock_fd)
            # One of the reason on why a socket could cause an error
            # is that the socket is already closed, ignore the
            # socket error if we try to close it at this point.
            with suppress(OSError):
                conn.close()

    def _from_server_socket(self, server_socket):  # noqa: C901  # FIXME
        try:
            s, addr = server_socket.accept()
            if self.server.stats['Enabled']:
                self.server.stats['Accepts'] += 1
            prevent_socket_inheritance(s)
            if hasattr(s, 'settimeout'):
                s.settimeout(self.server.timeout)

            mf = MakeFile
            ssl_env = {}
            # if ssl cert and key are set, we try to be a secure HTTP server
            if self.server.ssl_adapter is not None:
                try:
                    s, ssl_env = self.server.ssl_adapter.wrap(s)
                except errors.NoSSLError:
                    msg = (
                        'The client sent a plain HTTP request, but '
                        'this server only speaks HTTPS on this port.'
                    )
                    buf = [
                        '%s 400 Bad Request\r\n' % self.server.protocol,
                        'Content-Length: %s\r\n' % len(msg),
                        'Content-Type: text/plain\r\n\r\n',
                        msg,
                    ]

                    wfile = mf(s, 'wb', io.DEFAULT_BUFFER_SIZE)
                    try:
                        wfile.write(''.join(buf).encode('ISO-8859-1'))
                    except OSError as ex:
                        if ex.args[0] not in errors.socket_errors_to_ignore:
                            raise
                    return
                if not s:
                    return
                mf = self.server.ssl_adapter.makefile
                # Re-apply our timeout since we may have a new socket object
                if hasattr(s, 'settimeout'):
                    s.settimeout(self.server.timeout)

            conn = self.server.ConnectionClass(self.server, s, mf)

            if not isinstance(self.server.bind_addr, (str, bytes)):
                # optional values
                # Until we do DNS lookups, omit REMOTE_HOST
                if addr is None:  # sometimes this can happen
                    # figure out if AF_INET or AF_INET6.
                    if len(s.getsockname()) == 2:
                        # AF_INET
                        addr = ('0.0.0.0', 0)
                    else:
                        # AF_INET6
                        addr = ('::', 0)
                conn.remote_addr = addr[0]
                conn.remote_port = addr[1]

            conn.ssl_env = ssl_env
            return conn

        except socket.timeout:
            # The only reason for the timeout in start() is so we can
            # notice keyboard interrupts on Win32, which don't interrupt
            # accept() by default
            return
        except OSError as ex:
            if self.server.stats['Enabled']:
                self.server.stats['Socket Errors'] += 1
            if ex.args[0] in errors.socket_error_eintr:
                # I *think* this is right. EINTR should occur when a signal
                # is received during the accept() call; all docs say retry
                # the call, and I *think* I'm reading it right that Python
                # will then go ahead and poll for and handle the signal
                # elsewhere. See
                # https://github.com/cherrypy/cherrypy/issues/707.
                return
            if ex.args[0] in errors.socket_errors_nonblocking:
                # Just try again. See
                # https://github.com/cherrypy/cherrypy/issues/479.
                return
            if ex.args[0] in errors.socket_errors_to_ignore:
                # Our socket was closed.
                # See https://github.com/cherrypy/cherrypy/issues/686.
                return
            raise

    def close(self):
        """Close all monitored connections."""
        for (_, conn) in self._selector.connections:
            if conn is not self.server:  # server closes its own socket
                conn.close()
        self._selector.close()

    @property
    def _num_connections(self):
        """Return the current number of connections.

        Includes all connections registered with the selector,
        minus one for the server socket, which is always registered
        with the selector.
        """
        return len(self._selector) - 1

    @property
    def can_add_keepalive_connection(self):
        """Flag whether it is allowed to add a new keep-alive connection."""
        ka_limit = self.server.keep_alive_conn_limit
        return ka_limit is None or self._num_connections < ka_limit



"""
Manage Parallels Desktop VMs with ``prlctl`` and ``prlsrvctl``.  Only some of
the prlctl commands implemented so far.  Of those that have been implemented,
not all of the options may have been provided yet.  For a complete reference,
see the `Parallels Desktop Reference Guide
<http://download.parallels.com/desktop/v9/ga/docs/en_US/Parallels%20Command%20Line%20Reference%20Guide.pdf>`_.

This module requires the prlctl binary to be installed to run most functions.
To run parallels.prlsrvctl, the prlsrvctl binary is required.

What has not been implemented yet can be accessed through ``parallels.prlctl``
and ``parallels.prlsrvctl`` (note the preceding double dash ``--`` as
necessary):

.. code-block:: bash

    salt '*' parallels.prlctl installtools macvm runas=macdev
    salt -- '*' parallels.prlctl capture 'macvm --file macvm.display.png' runas=macdev
    salt -- '*' parallels.prlsrvctl set '--mem-limit auto' runas=macdev

.. versionadded:: 2016.3.0
"""

import logging
import re
import shlex

import salt.utils.data
import salt.utils.path
import salt.utils.yaml
from salt.exceptions import CommandExecutionError, SaltInvocationError

__virtualname__ = "parallels"
__func_alias__ = {
    "exec_": "exec",
}
log = logging.getLogger(__name__)
# Match any GUID
GUID_REGEX = re.compile(
    r"{?([0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})}?", re.I
)


def _normalize_args(args):
    """
    Return args as a list of strings
    """
    if isinstance(args, str):
        return shlex.split(args)

    if isinstance(args, (tuple, list)):
        return [str(arg) for arg in args]
    else:
        return [str(args)]


def _find_guids(guid_string):
    """
    Return the set of GUIDs found in guid_string

    :param str guid_string:
        String containing zero or more GUIDs.  Each GUID may or may not be
        enclosed in {}

    Example data (this string contains two distinct GUIDs):

    PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                            {a5b8999f-5d95-4aff-82de-e515b0101b66}
    {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
    """
    guids = []
    for found_guid in re.finditer(GUID_REGEX, guid_string):
        if found_guid.groups():
            guids.append(found_guid.group(0).strip("{}"))
    return sorted(list(set(guids)))


def prlsrvctl(sub_cmd, args=None, runas=None):
    """
    Execute a prlsrvctl command

    .. versionadded:: 2016.11.0

    :param str sub_cmd:
        prlsrvctl subcommand to execute

    :param str args:
        The arguments supplied to ``prlsrvctl <sub_cmd>``

    :param str runas:
        The user that the prlsrvctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.prlsrvctl info runas=macdev
        salt '*' parallels.prlsrvctl usb list runas=macdev
        salt -- '*' parallels.prlsrvctl set '--mem-limit auto' runas=macdev
    """
    if not salt.utils.path.which("prlsrvctl"):
        raise CommandExecutionError("prlsrvctl utility not available")

    # Construct command
    cmd = ["prlsrvctl", sub_cmd]
    if args:
        cmd.extend(_normalize_args(args))

    # Execute command and return output
    return __salt__["cmd.run"](cmd, runas=runas)


def prlctl(sub_cmd, args=None, runas=None):
    """
    Execute a prlctl command

    :param str sub_cmd:
        prlctl subcommand to execute

    :param str args:
        The arguments supplied to ``prlctl <sub_cmd>``

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.prlctl user list runas=macdev
        salt '*' parallels.prlctl exec 'macvm uname' runas=macdev
        salt -- '*' parallels.prlctl capture 'macvm --file macvm.display.png' runas=macdev
    """
    if not salt.utils.path.which("prlctl"):
        raise CommandExecutionError("prlctl utility not available")

    # Construct command
    cmd = ["prlctl", sub_cmd]
    if args:
        cmd.extend(_normalize_args(args))

    # Execute command and return output
    return __salt__["cmd.run"](cmd, runas=runas)


def list_vms(name=None, info=False, all=False, args=None, runas=None, template=False):
    """
    List information about the VMs

    :param str name:
        Name/ID of VM to list

        .. versionchanged:: 2016.11.0

            No longer implies ``info=True``

    :param str info:
        List extra information

    :param bool all:
        List all non-template VMs

    :param tuple args:
        Additional arguments given to ``prctl list``

    :param str runas:
        The user that the prlctl command will be run as

    :param bool template:
        List the available virtual machine templates.  The real virtual
        machines will not be included in the output

        .. versionadded:: 2016.11.0

    Example:

    .. code-block:: bash

        salt '*' parallels.list_vms runas=macdev
        salt '*' parallels.list_vms name=macvm info=True runas=macdev
        salt '*' parallels.list_vms info=True runas=macdev
        salt '*' parallels.list_vms ' -o uuid,status' all=True runas=macdev
    """
    # Construct argument list
    if args is None:
        args = []
    else:
        args = _normalize_args(args)

    if name:
        args.extend([name])
    if info:
        args.append("--info")
    if all:
        args.append("--all")
    if template:
        args.append("--template")

    # Execute command and return output
    return prlctl("list", args, runas=runas)


def clone(name, new_name, linked=False, template=False, runas=None):
    """
    Clone a VM

    .. versionadded:: 2016.11.0

    :param str name:
        Name/ID of VM to clone

    :param str new_name:
        Name of the new VM

    :param bool linked:
        Create a linked virtual machine.

    :param bool template:
        Create a virtual machine template instead of a real virtual machine.

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.clone macvm macvm_new runas=macdev
        salt '*' parallels.clone macvm macvm_templ template=True runas=macdev
    """
    args = [salt.utils.data.decode(name), "--name", salt.utils.data.decode(new_name)]
    if linked:
        args.append("--linked")
    if template:
        args.append("--template")
    return prlctl("clone", args, runas=runas)


def delete(name, runas=None):
    """
    Delete a VM

    .. versionadded:: 2016.11.0

    :param str name:
        Name/ID of VM to clone

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.exec macvm 'find /etc/paths.d' runas=macdev
    """
    return prlctl("delete", salt.utils.data.decode(name), runas=runas)


def exists(name, runas=None):
    """
    Query whether a VM exists

    .. versionadded:: 2016.11.0

    :param str name:
        Name/ID of VM

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.exists macvm runas=macdev
    """
    vm_info = list_vms(name, info=True, runas=runas).splitlines()
    for info_line in vm_info:
        if f"Name: {name}" in info_line:
            return True
    return False


def start(name, runas=None):
    """
    Start a VM

    :param str name:
        Name/ID of VM to start

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.start macvm runas=macdev
    """
    return prlctl("start", salt.utils.data.decode(name), runas=runas)


def stop(name, kill=False, runas=None):
    """
    Stop a VM

    :param str name:
        Name/ID of VM to stop

    :param bool kill:
        Perform a hard shutdown

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.stop macvm runas=macdev
        salt '*' parallels.stop macvm kill=True runas=macdev
    """
    # Construct argument list
    args = [salt.utils.data.decode(name)]
    if kill:
        args.append("--kill")

    # Execute command and return output
    return prlctl("stop", args, runas=runas)


def restart(name, runas=None):
    """
    Restart a VM by gracefully shutting it down and then restarting
    it

    :param str name:
        Name/ID of VM to restart

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.restart macvm runas=macdev
    """
    return prlctl("restart", salt.utils.data.decode(name), runas=runas)


def reset(name, runas=None):
    """
    Reset a VM by performing a hard shutdown and then a restart

    :param str name:
        Name/ID of VM to reset

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.reset macvm runas=macdev
    """
    return prlctl("reset", salt.utils.data.decode(name), runas=runas)


def status(name, runas=None):
    """
    Status of a VM

    :param str name:
        Name/ID of VM whose status will be returned

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.status macvm runas=macdev
    """
    return prlctl("status", salt.utils.data.decode(name), runas=runas)


def exec_(name, command, runas=None):
    """
    Run a command on a VM

    :param str name:
        Name/ID of VM whose exec will be returned

    :param str command:
        Command to run on the VM

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.exec macvm 'find /etc/paths.d' runas=macdev
    """
    # Construct argument list
    args = [salt.utils.data.decode(name)]
    args.extend(_normalize_args(command))

    # Execute command and return output
    return prlctl("exec", args, runas=runas)


def snapshot_id_to_name(name, snap_id, strict=False, runas=None):
    """
    Attempt to convert a snapshot ID to a snapshot name.  If the snapshot has
    no name or if the ID is not found or invalid, an empty string will be returned

    :param str name:
        Name/ID of VM whose snapshots are inspected

    :param str snap_id:
        ID of the snapshot

    :param bool strict:
        Raise an exception if a name cannot be found for the given ``snap_id``

    :param str runas:
        The user that the prlctl command will be run as

    Example data

    .. code-block:: yaml

        ID: {a5b8999f-5d95-4aff-82de-e515b0101b66}
        Name: original
        Date: 2016-03-04 10:50:34
        Current: yes
        State: poweroff
        Description: original state

    CLI Example:

    .. code-block:: bash

        salt '*' parallels.snapshot_id_to_name macvm a5b8999f-5d95-4aff-82de-e515b0101b66 runas=macdev
    """
    # Validate VM name and snapshot ID
    name = salt.utils.data.decode(name)
    if not re.match(GUID_REGEX, snap_id):
        raise SaltInvocationError(
            f'Snapshot ID "{salt.utils.data.decode(snap_id)}" is not a GUID'
        )

    # Get the snapshot information of the snapshot having the requested ID
    info = prlctl("snapshot-list", [name, "--id", snap_id], runas=runas)

    # Parallels desktop returned no information for snap_id
    if not info:
        raise SaltInvocationError(f'No snapshots for VM "{name}" have ID "{snap_id}"')

    # Try to interpret the information
    try:
        data = salt.utils.yaml.safe_load(info)
    except salt.utils.yaml.YAMLError as err:
        log.warning("Could not interpret snapshot data returned from prlctl: %s", err)
        data = {}

    # Find the snapshot name
    if isinstance(data, dict):
        snap_name = data.get("Name", "")
        # If snapshot name is of type NoneType, then the snapshot is unnamed
        if snap_name is None:
            snap_name = ""
    else:
        log.warning(
            "Could not interpret snapshot data returned from prlctl: "
            "data is not formed as a dictionary: %s",
            data,
        )
        snap_name = ""

    # Raise or return the result
    if not snap_name and strict:
        raise SaltInvocationError(
            'Could not find a snapshot name for snapshot ID "{}" of VM "{}"'.format(
                snap_id, name
            )
        )
    return salt.utils.data.decode(snap_name)


def snapshot_name_to_id(name, snap_name, strict=False, runas=None):
    """
    Attempt to convert a snapshot name to a snapshot ID.  If the name is not
    found an empty string is returned.  If multiple snapshots share the same
    name, a list will be returned

    :param str name:
        Name/ID of VM whose snapshots are inspected

    :param str snap_name:
        Name of the snapshot

    :param bool strict:
        Raise an exception if multiple snapshot IDs are found

    :param str runas:
        The user that the prlctl command will be run as

    CLI Example:

    .. code-block:: bash

        salt '*' parallels.snapshot_id_to_name macvm original runas=macdev
    """
    # Validate VM and snapshot names
    name = salt.utils.data.decode(name)
    snap_name = salt.utils.data.decode(snap_name)

    # Get a multiline string containing all the snapshot GUIDs
    info = prlctl("snapshot-list", name, runas=runas)

    # Get a set of all snapshot GUIDs in the string
    snap_ids = _find_guids(info)

    # Try to match the snapshot name to an ID
    named_ids = []
    for snap_id in snap_ids:
        if snapshot_id_to_name(name, snap_id, runas=runas) == snap_name:
            named_ids.append(snap_id)

    # Return one or more IDs having snap_name or raise an error upon
    # non-singular names
    if not named_ids:
        raise SaltInvocationError(
            f'No snapshots for VM "{name}" have name "{snap_name}"'
        )
    elif len(named_ids) == 1:
        return named_ids[0]
    else:
        multi_msg = 'Multiple snapshots for VM "{}" have name "{}"'.format(
            name, snap_name
        )
        if strict:
            raise SaltInvocationError(multi_msg)
        else:
            log.warning(multi_msg)
        return named_ids


def _validate_snap_name(name, snap_name, strict=True, runas=None):
    """
    Validate snapshot name and convert to snapshot ID

    :param str name:
        Name/ID of VM whose snapshot name is being validated

    :param str snap_name:
        Name/ID of snapshot

    :param bool strict:
        Raise an exception if multiple snapshot IDs are found

    :param str runas:
        The user that the prlctl command will be run as
    """
    snap_name = salt.utils.data.decode(snap_name)

    # Try to convert snapshot name to an ID without {}
    if re.match(GUID_REGEX, snap_name):
        return snap_name.strip("{}")
    else:
        return snapshot_name_to_id(name, snap_name, strict=strict, runas=runas)


def list_snapshots(name, snap_name=None, tree=False, names=False, runas=None):
    """
    List the snapshots

    :param str name:
        Name/ID of VM whose snapshots will be listed

    :param str snap_id:
        Name/ID of snapshot to display information about.  If ``tree=True`` is
        also specified, display the snapshot subtree having this snapshot as
        the root snapshot

    :param bool tree:
        List snapshots in tree format rather than tabular format

    :param bool names:
        List snapshots as ID, name pairs

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.list_snapshots macvm runas=macdev
        salt '*' parallels.list_snapshots macvm tree=True runas=macdev
        salt '*' parallels.list_snapshots macvm snap_name=original runas=macdev
        salt '*' parallels.list_snapshots macvm names=True runas=macdev
    """
    # Validate VM and snapshot names
    name = salt.utils.data.decode(name)
    if snap_name:
        snap_name = _validate_snap_name(name, snap_name, runas=runas)

    # Construct argument list
    args = [name]
    if tree:
        args.append("--tree")
    if snap_name:
        args.extend(["--id", snap_name])

    # Execute command
    res = prlctl("snapshot-list", args, runas=runas)

    # Construct ID, name pairs
    if names:
        # Find all GUIDs in the result
        snap_ids = _find_guids(res)

        # Try to find the snapshot names
        ret = "{:<38}  {}\n".format("Snapshot ID", "Snapshot Name")
        for snap_id in snap_ids:
            snap_name = snapshot_id_to_name(name, snap_id, runas=runas)
            ret += f"{{{snap_id}}}  {salt.utils.data.decode(snap_name)}\n"
        return ret

    # Return information directly from parallels desktop
    else:
        return res


def snapshot(name, snap_name=None, desc=None, runas=None):
    """
    Create a snapshot

    :param str name:
        Name/ID of VM to take a snapshot of

    :param str snap_name:
        Name of snapshot

    :param str desc:
        Description of snapshot

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.create_snapshot macvm snap_name=macvm-original runas=macdev
        salt '*' parallels.create_snapshot macvm snap_name=macvm-updates desc='clean install with updates' runas=macdev
    """
    # Validate VM and snapshot names
    name = salt.utils.data.decode(name)
    if snap_name:
        snap_name = salt.utils.data.decode(snap_name)

    # Construct argument list
    args = [name]
    if snap_name:
        args.extend(["--name", snap_name])
    if desc:
        args.extend(["--description", desc])

    # Execute command and return output
    return prlctl("snapshot", args, runas=runas)


def delete_snapshot(name, snap_name, runas=None, all=False):
    """
    Delete a snapshot

    .. note::

        Deleting a snapshot from which other snapshots are dervied will not
        delete the derived snapshots

    :param str name:
        Name/ID of VM whose snapshot will be deleted

    :param str snap_name:
        Name/ID of snapshot to delete

    :param str runas:
        The user that the prlctl command will be run as

    :param bool all:
        Delete all snapshots having the name given

        .. versionadded:: 2016.11.0

    Example:

    .. code-block:: bash

        salt '*' parallels.delete_snapshot macvm 'unneeded snapshot' runas=macdev
        salt '*' parallels.delete_snapshot macvm 'Snapshot for linked clone' all=True runas=macdev
    """
    # strict means raise an error if multiple snapshot IDs found for the name given
    strict = not all

    # Validate VM and snapshot names
    name = salt.utils.data.decode(name)
    snap_ids = _validate_snap_name(name, snap_name, strict=strict, runas=runas)
    if isinstance(snap_ids, str):
        snap_ids = [snap_ids]

    # Delete snapshot(s)
    ret = {}
    for snap_id in snap_ids:
        snap_id = snap_id.strip("{}")
        # Construct argument list
        args = [name, "--id", snap_id]

        # Execute command
        ret[snap_id] = prlctl("snapshot-delete", args, runas=runas)

    # Return results
    ret_keys = list(ret.keys())
    if len(ret_keys) == 1:
        return ret[ret_keys[0]]
    else:
        return ret


def revert_snapshot(name, snap_name, runas=None):
    """
    Revert a VM to a snapshot

    :param str name:
        Name/ID of VM to revert to a snapshot

    :param str snap_name:
        Name/ID of snapshot to revert to

    :param str runas:
        The user that the prlctl command will be run as

    Example:

    .. code-block:: bash

        salt '*' parallels.revert_snapshot macvm base-with-updates runas=macdev
    """
    # Validate VM and snapshot names
    name = salt.utils.data.decode(name)
    snap_name = _validate_snap_name(name, snap_name, runas=runas)

    # Construct argument list
    args = [name, "--id", snap_name]

    # Execute command and return output
    return prlctl("snapshot-switch", args, runas=runas)



"""
QingCloud Cloud Module
======================

.. versionadded:: 2015.8.0

The QingCloud cloud module is used to control access to the QingCloud.
http://www.qingcloud.com/

Use of this module requires the ``access_key_id``, ``secret_access_key``,
``zone`` and ``key_filename`` parameter to be set.

Set up the cloud configuration at ``/etc/salt/cloud.providers`` or
``/etc/salt/cloud.providers.d/qingcloud.conf``:

.. code-block:: yaml

    my-qingcloud:
      driver: qingcloud
      access_key_id: AKIDMRTGYONNLTFFRBQJ
      secret_access_key: clYwH21U5UOmcov4aNV2V2XocaHCG3JZGcxEczFu
      zone: pek2
      key_filename: /path/to/your.pem

:depends: requests
"""

import base64
import hmac
import logging
import pprint
import time
import urllib.parse
from hashlib import sha256

import salt.config as config
import salt.utils.cloud
import salt.utils.data
import salt.utils.json
from salt.exceptions import (
    SaltCloudExecutionFailure,
    SaltCloudExecutionTimeout,
    SaltCloudNotFound,
    SaltCloudSystemExit,
)

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False


# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "qingcloud"

DEFAULT_QINGCLOUD_API_VERSION = 1
DEFAULT_QINGCLOUD_SIGNATURE_VERSION = 1


# Only load in this module if the qingcloud configurations are in place
def __virtual__():
    """
    Check for QingCloud configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or __virtualname__,
        ("access_key_id", "secret_access_key", "zone", "key_filename"),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"requests": HAS_REQUESTS})


def _compute_signature(parameters, access_key_secret, method, path):
    """
    Generate an API request signature. Detailed document can be found at:

    https://docs.qingcloud.com/api/common/signature.html
    """
    parameters["signature_method"] = "HmacSHA256"

    string_to_sign = f"{method.upper()}\n{path}\n"

    keys = sorted(parameters.keys())
    pairs = []
    for key in keys:
        val = str(parameters[key]).encode("utf-8")
        pairs.append(
            urllib.parse.quote(key, safe="") + "=" + urllib.parse.quote(val, safe="-_~")
        )
    qs = "&".join(pairs)
    string_to_sign += qs

    h = hmac.new(access_key_secret, digestmod=sha256)
    h.update(string_to_sign)

    signature = base64.b64encode(h.digest()).strip()

    return signature


def query(params=None):
    """
    Make a web call to QingCloud IaaS API.
    """
    path = "https://api.qingcloud.com/iaas/"

    access_key_id = config.get_cloud_config_value(
        "access_key_id", get_configured_provider(), __opts__, search_global=False
    )
    access_key_secret = config.get_cloud_config_value(
        "secret_access_key", get_configured_provider(), __opts__, search_global=False
    )

    verify_ssl = config.get_cloud_config_value(
        "verify_ssl",
        get_configured_provider(),
        __opts__,
        default=True,
        search_global=False,
    )

    # public interface parameters
    real_parameters = {
        "access_key_id": access_key_id,
        "signature_version": DEFAULT_QINGCLOUD_SIGNATURE_VERSION,
        "time_stamp": time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime()),
        "version": DEFAULT_QINGCLOUD_API_VERSION,
    }

    # include action or function parameters
    if params:
        for key, value in params.items():
            if isinstance(value, list):
                for i in range(1, len(value) + 1):
                    if isinstance(value[i - 1], dict):
                        for sk, sv in value[i - 1].items():
                            if isinstance(sv, dict) or isinstance(sv, list):
                                sv = salt.utils.json.dumps(sv, separators=(",", ":"))
                            real_parameters[f"{key}.{i}.{sk}"] = sv
                    else:
                        real_parameters[f"{key}.{i}"] = value[i - 1]
            else:
                real_parameters[key] = value

    # Calculate the string for Signature
    signature = _compute_signature(real_parameters, access_key_secret, "GET", "/iaas/")
    real_parameters["signature"] = signature

    # print('parameters:')
    # pprint.pprint(real_parameters)

    request = requests.get(path, params=real_parameters, verify=verify_ssl, timeout=120)

    # print('url:')
    # print(request.url)

    if request.status_code != 200:
        raise SaltCloudSystemExit(
            "An error occurred while querying QingCloud. HTTP Code: {}  "
            "Error: '{}'".format(request.status_code, request.text)
        )

    log.debug(request.url)

    content = request.text
    result = salt.utils.json.loads(content)

    # print('response:')
    # pprint.pprint(result)

    if result["ret_code"] != 0:
        raise SaltCloudSystemExit(pprint.pformat(result.get("message", {})))

    return result


def avail_locations(call=None):
    """
    Return a dict of all available locations on the provider with
    relevant data.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-locations my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_locations function must be called with "
            "-f or --function, or with the --list-locations option"
        )

    params = {
        "action": "DescribeZones",
    }
    items = query(params=params)

    result = {}
    for region in items["zone_set"]:
        result[region["zone_id"]] = {}
        for key in region:
            result[region["zone_id"]][key] = str(region[key])

    return result


def _get_location(vm_=None):
    """
    Return the VM's location. Used by create().
    """
    locations = avail_locations()

    vm_location = str(
        config.get_cloud_config_value("zone", vm_, __opts__, search_global=False)
    )

    if not vm_location:
        raise SaltCloudNotFound("No location specified for this VM.")

    if vm_location in locations:
        return vm_location

    raise SaltCloudNotFound(
        f"The specified location, '{vm_location}', could not be found."
    )


def _get_specified_zone(kwargs=None, provider=None):
    if provider is None:
        provider = get_configured_provider()

    if isinstance(kwargs, dict):
        zone = kwargs.get("zone", None)
        if zone is not None:
            return zone

    zone = provider["zone"]
    return zone


def avail_images(kwargs=None, call=None):
    """
    Return a list of the images that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-images my-qingcloud
        salt-cloud -f avail_images my-qingcloud zone=gd1
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with "
            "-f or --function, or with the --list-images option"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    params = {
        "action": "DescribeImages",
        "provider": "system",
        "zone": _get_specified_zone(kwargs, get_configured_provider()),
    }
    items = query(params=params)

    result = {}
    for image in items["image_set"]:
        result[image["image_id"]] = {}
        for key in image:
            result[image["image_id"]][key] = image[key]

    return result


def _get_image(vm_):
    """
    Return the VM's image. Used by create().
    """
    images = avail_images()
    vm_image = str(
        config.get_cloud_config_value("image", vm_, __opts__, search_global=False)
    )

    if not vm_image:
        raise SaltCloudNotFound("No image specified for this VM.")

    if vm_image in images:
        return vm_image

    raise SaltCloudNotFound(f"The specified image, '{vm_image}', could not be found.")


def show_image(kwargs, call=None):
    """
    Show the details from QingCloud concerning an image.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f show_image my-qingcloud image=trustysrvx64c
        salt-cloud -f show_image my-qingcloud image=trustysrvx64c,coreos4
        salt-cloud -f show_image my-qingcloud image=trustysrvx64c zone=ap1
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The show_images function must be called with -f or --function"
        )

    if not isinstance(kwargs, dict):
        kwargs = {}

    images = kwargs["image"]
    images = images.split(",")

    params = {
        "action": "DescribeImages",
        "images": images,
        "zone": _get_specified_zone(kwargs, get_configured_provider()),
    }

    items = query(params=params)

    if not items["image_set"]:
        raise SaltCloudNotFound("The specified image could not be found.")

    result = {}
    for image in items["image_set"]:
        result[image["image_id"]] = {}
        for key in image:
            result[image["image_id"]][key] = image[key]

    return result


# QingCloud doesn't provide an API of geting instance sizes
QINGCLOUD_SIZES = {
    "pek2": {
        "c1m1": {"cpu": 1, "memory": "1G"},
        "c1m2": {"cpu": 1, "memory": "2G"},
        "c1m4": {"cpu": 1, "memory": "4G"},
        "c2m2": {"cpu": 2, "memory": "2G"},
        "c2m4": {"cpu": 2, "memory": "4G"},
        "c2m8": {"cpu": 2, "memory": "8G"},
        "c4m4": {"cpu": 4, "memory": "4G"},
        "c4m8": {"cpu": 4, "memory": "8G"},
        "c4m16": {"cpu": 4, "memory": "16G"},
    },
    "pek1": {
        "small_b": {"cpu": 1, "memory": "1G"},
        "small_c": {"cpu": 1, "memory": "2G"},
        "medium_a": {"cpu": 2, "memory": "2G"},
        "medium_b": {"cpu": 2, "memory": "4G"},
        "medium_c": {"cpu": 2, "memory": "8G"},
        "large_a": {"cpu": 4, "memory": "4G"},
        "large_b": {"cpu": 4, "memory": "8G"},
        "large_c": {"cpu": 4, "memory": "16G"},
    },
}
QINGCLOUD_SIZES["ap1"] = QINGCLOUD_SIZES["pek2"]
QINGCLOUD_SIZES["gd1"] = QINGCLOUD_SIZES["pek2"]


def avail_sizes(kwargs=None, call=None):
    """
    Return a list of the instance sizes that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud --list-sizes my-qingcloud
        salt-cloud -f avail_sizes my-qingcloud zone=pek2
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with "
            "-f or --function, or with the --list-sizes option"
        )

    zone = _get_specified_zone(kwargs, get_configured_provider())

    result = {}
    for size_key in QINGCLOUD_SIZES[zone]:
        result[size_key] = {}
        for attribute_key in QINGCLOUD_SIZES[zone][size_key]:
            result[size_key][attribute_key] = QINGCLOUD_SIZES[zone][size_key][
                attribute_key
            ]

    return result


def _get_size(vm_):
    """
    Return the VM's size. Used by create().
    """
    sizes = avail_sizes()

    vm_size = str(
        config.get_cloud_config_value("size", vm_, __opts__, search_global=False)
    )

    if not vm_size:
        raise SaltCloudNotFound("No size specified for this instance.")

    if vm_size in sizes:
        return vm_size

    raise SaltCloudNotFound(f"The specified size, '{vm_size}', could not be found.")


def _show_normalized_node(full_node):
    """
    Normalize the QingCloud instance data. Used by list_nodes()-related
    functions.
    """
    public_ips = full_node.get("eip", [])
    if public_ips:
        public_ip = public_ips["eip_addr"]
        public_ips = [
            public_ip,
        ]

    private_ips = []
    for vxnet in full_node.get("vxnets", []):
        private_ip = vxnet.get("private_ip", None)
        if private_ip:
            private_ips.append(private_ip)

    normalized_node = {
        "id": full_node["instance_id"],
        "image": full_node["image"]["image_id"],
        "size": full_node["instance_type"],
        "state": full_node["status"],
        "private_ips": private_ips,
        "public_ips": public_ips,
    }

    return normalized_node


def list_nodes_full(call=None):
    """
    Return a list of the instances that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -F my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function."
        )

    zone = _get_specified_zone()

    params = {
        "action": "DescribeInstances",
        "zone": zone,
        "status": ["pending", "running", "stopped", "suspended"],
    }
    items = query(params=params)

    log.debug("Total %s instances found in zone %s", items["total_count"], zone)

    result = {}

    if items["total_count"] == 0:
        return result

    for node in items["instance_set"]:
        normalized_node = _show_normalized_node(node)
        node.update(normalized_node)

        result[node["instance_id"]] = node

    provider = _get_active_provider_name() or "qingcloud"
    if ":" in provider:
        comps = provider.split(":")
        provider = comps[0]

    __opts__["update_cachedir"] = True
    __utils__["cloud.cache_node_list"](result, provider, __opts__)

    return result


def list_nodes(call=None):
    """
    Return a list of the instances that are on the provider.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -Q my-qingcloud
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function."
        )

    nodes = list_nodes_full()

    ret = {}
    for instance_id, full_node in nodes.items():
        ret[instance_id] = {
            "id": full_node["id"],
            "image": full_node["image"],
            "size": full_node["size"],
            "state": full_node["state"],
            "public_ips": full_node["public_ips"],
            "private_ips": full_node["private_ips"],
        }

    return ret


def list_nodes_min(call=None):
    """
    Return a list of the instances that are on the provider. Only a list of
    instances names, and their state, is returned.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -f list_nodes_min my-qingcloud
    """
    if call != "function":
        raise SaltCloudSystemExit(
            "The list_nodes_min function must be called with -f or --function."
        )

    nodes = list_nodes_full()

    result = {}
    for instance_id, full_node in nodes.items():
        result[instance_id] = {
            "name": full_node["instance_name"],
            "status": full_node["status"],
        }

    return result


def list_nodes_select(call=None):
    """
    Return a list of the instances that are on the provider, with selected
    fields.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -S my-qingcloud
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"),
        __opts__["query.selection"],
        call,
    )


def show_instance(instance_id, call=None, kwargs=None):
    """
    Show the details from QingCloud concerning an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a show_instance i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance action must be called with -a or --action."
        )

    params = {
        "action": "DescribeInstances",
        "instances.1": instance_id,
        "zone": _get_specified_zone(kwargs=None, provider=get_configured_provider()),
    }
    items = query(params=params)

    if items["total_count"] == 0:
        raise SaltCloudNotFound(
            f"The specified instance, '{instance_id}', could not be found."
        )

    full_node = items["instance_set"][0]
    normalized_node = _show_normalized_node(full_node)
    full_node.update(normalized_node)

    result = full_node

    return result


def _query_node_data(instance_id):
    data = show_instance(instance_id, call="action")

    if not data:
        return False

    if data.get("private_ips", []):
        return data


def create(vm_):
    """
    Create a single instance from a data dict.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -p qingcloud-ubuntu-c1m1 hostname1
        salt-cloud -m /path/to/mymap.sls -P
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "qingcloud",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])

    # params
    params = {
        "action": "RunInstances",
        "instance_name": vm_["name"],
        "zone": _get_location(vm_),
        "instance_type": _get_size(vm_),
        "image_id": _get_image(vm_),
        "vxnets.1": vm_["vxnets"],
        "login_mode": vm_["login_mode"],
        "login_keypair": vm_["login_keypair"],
    }

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        args={
            "kwargs": __utils__["cloud.filter_event"](
                "requesting", params, list(params)
            ),
        },
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    result = query(params)
    new_instance_id = result["instances"][0]

    try:
        data = salt.utils.cloud.wait_for_ip(
            _query_node_data,
            update_args=(new_instance_id,),
            timeout=config.get_cloud_config_value(
                "wait_for_ip_timeout", vm_, __opts__, default=10 * 60
            ),
            interval=config.get_cloud_config_value(
                "wait_for_ip_interval", vm_, __opts__, default=10
            ),
        )
    except (SaltCloudExecutionTimeout, SaltCloudExecutionFailure) as exc:
        try:
            # It might be already up, let's destroy it!
            destroy(vm_["name"])
        except SaltCloudSystemExit:
            pass
        finally:
            raise SaltCloudSystemExit(str(exc))

    private_ip = data["private_ips"][0]

    log.debug("VM %s is now running", private_ip)

    vm_["ssh_host"] = private_ip

    # The instance is booted and accessible, let's Salt it!
    __utils__["cloud.bootstrap"](vm_, __opts__)

    log.info("Created Cloud VM '%s'", vm_["name"])

    log.debug("'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data))

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return data


def script(vm_):
    """
    Return the script deployment object.
    """
    deploy_script = salt.utils.cloud.os_script(
        config.get_cloud_config_value("script", vm_, __opts__),
        vm_,
        __opts__,
        salt.utils.cloud.salt_config_to_yaml(
            salt.utils.cloud.minion_config(__opts__, vm_)
        ),
    )

    return deploy_script


def start(instance_id, call=None):
    """
    Start an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a start i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Starting instance %s", instance_id)

    params = {
        "action": "StartInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    return result


def stop(instance_id, force=False, call=None):
    """
    Stop an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a stop i-2f733r5n
        salt-cloud -a stop i-2f733r5n force=True
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Stopping instance %s", instance_id)

    params = {
        "action": "StopInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
        "force": int(force),
    }
    result = query(params)

    return result


def reboot(instance_id, call=None):
    """
    Reboot an instance.

    CLI Examples:

    .. code-block:: bash

        salt-cloud -a reboot i-2f733r5n
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    log.info("Rebooting instance %s", instance_id)

    params = {
        "action": "RestartInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    return result


def destroy(instance_id, call=None):
    """
    Destroy an instance.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a destroy i-2f733r5n
        salt-cloud -d i-2f733r5n
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    instance_data = show_instance(instance_id, call="action")
    name = instance_data["instance_name"]

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    params = {
        "action": "TerminateInstances",
        "zone": _get_specified_zone(provider=get_configured_provider()),
        "instances.1": instance_id,
    }
    result = query(params)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return result



"""
This is the default nodegroup matcher.
"""

import logging

import salt.loader
import salt.utils.minions

log = logging.getLogger(__name__)


def _load_matchers(opts):
    """
    Store matchers in __context__ so they're only loaded once
    """
    __context__["matchers"] = salt.loader.matchers(opts)


def match(tgt, nodegroups=None, opts=None, minion_id=None):
    """
    This is a compatibility matcher and is NOT called when using
    nodegroups for remote execution, but is called when the nodegroups
    matcher is used in states
    """
    if not opts:
        opts = __opts__
    if not nodegroups:
        log.debug("Nodegroup matcher called with no nodegroups.")
        return False
    if tgt in nodegroups:
        if "matchers" not in __context__:
            _load_matchers(opts)
        return __context__["matchers"]["compound_match.match"](
            salt.utils.minions.nodegroup_comp(tgt, nodegroups)
        )
    return False



# Copyright 2022-2025 Broadcom.
# SPDX-License-Identifier: Apache-2
"""
The ``relenv fetch`` command.
"""

import os
import sys

from .build import platform_module
from .common import (
    CHECK_HOSTS,
    DATA_DIR,
    DEFAULT_PYTHON,
    __version__,
    build_arch,
    check_url,
    download_url,
    get_triplet,
    work_dir,
)


def setup_parser(subparsers):
    """
    Setup the subparser for the ``fetch`` command.

    :param subparsers: The subparsers object returned from ``add_subparsers``
    :type subparsers: argparse._SubParsersAction
    """
    mod = platform_module()
    subparser = subparsers.add_parser("fetch", description="Fetch relenv builds")
    subparser.set_defaults(func=main)
    subparser.add_argument(
        "--arch",
        default=build_arch(),
        choices=mod.ARCHES,
        type=str,
        help="Architecture to download. [default: %(default)s]",
    )
    subparser.add_argument(
        "--python",
        default=DEFAULT_PYTHON,
        type=str,
        help="The python version [default: %(default)s]",
    )


def fetch(version, triplet, python, check_hosts=CHECK_HOSTS):
    """
    Fetch the specified python build.
    """
    url = f"https://github.com/saltstack/relenv/releases/download/v{version}/{python}-{triplet}.tar.xz"
    if not check_url(url, timeout=5):
        for host in check_hosts:
            url = f"https://{host}/relenv/{version}/build/{python}-{triplet}.tar.xz"
            if check_url(url, timeout=5):
                break
        else:
            print(
                f"Unable to find file on any hosts: github.com {' '.join(x.split('/')[0] for x in check_hosts)}"
            )
            sys.exit(1)
    builddir = work_dir("build", DATA_DIR)
    os.makedirs(builddir, exist_ok=True)
    download_url(url, builddir)


def main(args):
    """
    The entrypoint into the ``relenv fetch`` command.

    :param args: The args passed to the command
    :type args: argparse.Namespace
    """
    version = os.environ.get("RELENV_FETCH_VERSION", __version__)
    triplet = get_triplet(machine=args.arch)
    python = args.python
    check_hosts = CHECK_HOSTS
    if os.environ.get("RELENV_FETCH_HOST", ""):
        check_hosts = [os.environ["RELENV_FETCH_HOST"]]
    fetch(version, triplet, python, check_hosts)



"""
Runner to provide F5 Load Balancer functionality

:depends:   - pycontrol Python module

:configuration: In order to connect to a F5 Load Balancer, you must specify
    in the Salt master configuration the currently available load balancers

    .. code-block:: yaml

        load_balancers:
          bigip1.example.com:
            username: admin
            password: secret
          bigip2.example.com:
            username: admin
            password: secret
"""

from salt.exceptions import CommandExecutionError

try:
    import pycontrol.pycontrol as f5

    HAS_PYCONTROL = True
except ImportError:
    HAS_PYCONTROL = False


def __virtual__():
    if not HAS_PYCONTROL:
        return False
    return True


class F5Mgmt:
    def __init__(self, lb, username, password):
        self.lb = lb
        self.username = username
        self.password = password
        self._connect()

    def _connect(self):
        """
        Connect to F5
        """
        try:
            self.bigIP = f5.BIGIP(
                hostname=self.lb,
                username=self.username,
                password=self.password,
                fromurl=True,
                wsdls=["LocalLB.VirtualServer", "LocalLB.Pool"],
            )
        except Exception:  # pylint: disable=broad-except
            raise Exception(f"Unable to connect to {self.lb}")

        return True

    def create_vs(self, name, ip, port, protocol, profile, pool_name):
        """
        Create a virtual server
        """
        vs = self.bigIP.LocalLB.VirtualServer
        vs_def = vs.typefactory.create("Common.VirtualServerDefinition")

        vs_def.name = name
        vs_def.address = ip
        vs_def.port = port

        common_protocols = vs.typefactory.create("Common.ProtocolType")

        p = [i[0] for i in common_protocols if i[0].split("_")[1] == protocol.upper()]

        if p:
            vs_def.protocol = p
        else:
            raise CommandExecutionError("Unknown protocol")

        vs_def_seq = vs.typefactory.create("Common.VirtualServerSequence")
        vs_def_seq.item = [vs_def]

        vs_type = vs.typefactory.create("LocalLB.VirtualServer.VirtualServerType")
        vs_resource = vs.typefactory.create(
            "LocalLB.VirtualServer.VirtualServerResource"
        )

        vs_resource.type = vs_type.RESOURCE_TYPE_POOL
        vs_resource.default_pool_name = pool_name

        resource_seq = vs.typefactory.create(
            "LocalLB.VirtualServer.VirtualServerResourceSequence"
        )

        resource_seq.item = [vs_resource]

        vs_context = vs.typefactory.create("LocalLB.ProfileContextType")
        vs_profile = vs.typefactory.create("LocalLB.VirtualServer.VirtualServerProfile")

        vs_profile.profile_context = vs_context.PROFILE_CONTEXT_TYPE_ALL
        vs_profile.profile_name = protocol

        vs_profile_http = vs.typefactory.create(
            "LocalLB.VirtualServer.VirtualServerProfile"
        )
        vs_profile_http.profile_name = profile

        vs_profile_conn = vs.typefactory.create(
            "LocalLB.VirtualServer.VirtualServerProfile"
        )
        vs_profile_conn.profile_name = "oneconnect"

        vs_profile_seq = vs.typefactory.create(
            "LocalLB.VirtualServer.VirtualServerProfileSequence"
        )
        vs_profile_seq.item = [vs_profile, vs_profile_http, vs_profile_conn]

        try:
            vs.create(
                definitions=vs_def_seq,
                wildmasks=["255.255.255.255"],
                resources=resource_seq,
                profiles=[vs_profile_seq],
            )
        except Exception as e:  # pylint: disable=broad-except
            raise Exception(f"Unable to create `{name}` virtual server\n\n{e}")
        return True

    def create_pool(self, name, method="ROUND_ROBIN"):
        """
        Create a pool on the F5 load balancer
        """
        lbmethods = self.bigIP.LocalLB.Pool.typefactory.create("LocalLB.LBMethod")

        supported_method = [
            i[0] for i in lbmethods if (i[0].split("_", 2)[-1] == method.upper())
        ]

        if supported_method and not self.check_pool(name):
            try:
                self.bigIP.LocalLB.Pool.create(
                    pool_names=[name], lb_methods=[supported_method], members=[[]]
                )
            except Exception as e:  # pylint: disable=broad-except
                raise Exception(f"Unable to create `{name}` pool\n\n{e}")
        else:
            raise Exception("Unsupported method")
        return True

    def add_pool_member(self, name, port, pool_name):
        """
        Add a node to a pool
        """
        if not self.check_pool(pool_name):
            raise CommandExecutionError(f"{pool_name} pool does not exists")

        members_seq = self.bigIP.LocalLB.Pool.typefactory.create(
            "Common.IPPortDefinitionSequence"
        )
        members_seq.items = []

        member = self.bigIP.LocalLB.Pool.typefactory.create("Common.IPPortDefinition")

        member.address = name
        member.port = port

        members_seq.items.append(member)

        try:
            self.bigIP.LocalLB.Pool.add_member(
                pool_names=[pool_name], members=[members_seq]
            )
        except Exception as e:  # pylint: disable=broad-except
            raise Exception(f"Unable to add `{name}` to `{pool_name}`\n\n{e}")
        return True

    def check_pool(self, name):
        """
        Check to see if a pool exists
        """
        pools = self.bigIP.LocalLB.Pool
        for pool in pools.get_list():
            if pool.split("/")[-1] == name:
                return True
        return False

    def check_virtualserver(self, name):
        """
        Check to see if a virtual server exists
        """
        vs = self.bigIP.LocalLB.VirtualServer
        for v in vs.get_list():
            if v.split("/")[-1] == name:
                return True
        return False

    def check_member_pool(self, member, pool_name):
        """
        Check a pool member exists in a specific pool
        """
        members = self.bigIP.LocalLB.Pool.get_member(pool_names=[pool_name])[0]
        for mem in members:
            if member == mem.address:
                return True
        return False

    def lbmethods(self):
        """
        List all the load balancer methods
        """
        methods = self.bigIP.LocalLB.Pool.typefactory.create("LocalLB.LBMethod")
        return [method[0].split("_", 2)[-1] for method in methods]


def create_vs(lb, name, ip, port, protocol, profile, pool_name):
    """
    Create a virtual server

    CLI Examples:

    .. code-block:: bash

        salt-run f5.create_vs lbalancer vs_name 10.0.0.1 80 tcp http poolname
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")

    F5 = F5Mgmt(lb, username, password)
    F5.create_vs(name, ip, port, protocol, profile, pool_name)
    return True


def create_pool(lb, name, method="ROUND_ROBIN"):
    """
    Create a pool on the F5 load balancer

    CLI Examples:

    .. code-block:: bash

        salt-run f5.create_pool load_balancer pool_name loadbalance_method
        salt-run f5.create_pool load_balancer my_pool ROUND_ROBIN
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")
    F5 = F5Mgmt(lb, username, password)
    F5.create_pool(name, method)
    return True


def add_pool_member(lb, name, port, pool_name):
    """
    Add a node to a pool

    CLI Examples:

    .. code-block:: bash

        salt-run f5.add_pool_member load_balancer 10.0.0.1 80 my_pool
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")
    F5 = F5Mgmt(lb, username, password)
    F5.add_pool_member(name, port, pool_name)
    return True


def check_pool(lb, name):
    """
    Check to see if a pool exists

    CLI Examples:

    .. code-block:: bash

        salt-run f5.check_pool load_balancer pool_name
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")
    F5 = F5Mgmt(lb, username, password)
    return F5.check_pool(name)


def check_virtualserver(lb, name):
    """
    Check to see if a virtual server exists

    CLI Examples:

    .. code-block:: bash

        salt-run f5.check_virtualserver load_balancer virtual_server
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")
    F5 = F5Mgmt(lb, username, password)
    return F5.check_virtualserver(name)


def check_member_pool(lb, member, pool_name):
    """
    Check a pool member exists in a specific pool

    CLI Examples:

    .. code-block:: bash

        salt-run f5.check_member_pool load_balancer 10.0.0.1 my_pool
    """
    if __opts__["load_balancers"].get(lb, None):
        (username, password) = list(__opts__["load_balancers"][lb].values())
    else:
        raise Exception(f"Unable to find `{lb}` load balancer")
    F5 = F5Mgmt(lb, username, password)
    return F5.check_member_pool(member, pool_name)



"""
Module for apcupsd
"""

import logging

import salt.utils.decorators as decorators
import salt.utils.path

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "apcups"


@decorators.memoize
def _check_apcaccess():
    """
    Looks to see if apcaccess is present on the system
    """
    return salt.utils.path.which("apcaccess")


def __virtual__():
    """
    Provides apcupsd only if apcaccess is present
    """
    if _check_apcaccess():
        return __virtualname__
    return (
        False,
        "{} module can only be loaded on when apcupsd is installed".format(
            __virtualname__
        ),
    )


def status():
    """
    Return apcaccess output

    CLI Example:

    .. code-block:: bash

        salt '*' apcups.status
    """
    ret = {}
    apcaccess = _check_apcaccess()
    res = __salt__["cmd.run_all"](apcaccess)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = "Something with wrong executing apcaccess, is apcupsd running?"
        return ret

    for line in res["stdout"].splitlines():
        line = line.split(":")
        ret[line[0].strip()] = line[1].strip()

    return ret


def status_load():
    """
    Return load

    CLI Example:

    .. code-block:: bash

        salt '*' apcups.status_load
    """
    data = status()
    if "LOADPCT" in data:
        load = data["LOADPCT"].split()
        if load[1].lower() == "percent":
            return float(load[0])

    return {"Error": "Load not available."}


def status_charge():
    """
    Return battery charge

    CLI Example:

    .. code-block:: bash

        salt '*' apcups.status_charge
    """
    data = status()
    if "BCHARGE" in data:
        charge = data["BCHARGE"].split()
        if charge[1].lower() == "percent":
            return float(charge[0])

    return {"Error": "Load not available."}


def status_battery():
    """
    Return true if running on battery power

    CLI Example:

    .. code-block:: bash

        salt '*' apcups.status_battery
    """
    data = status()
    if "TONBATT" in data:
        return not data["TONBATT"] == "0 Seconds"

    return {"Error": "Battery status not available."}



"""
Beacon that fires events on vm state changes

.. code-block:: yaml

    ## minimal
    # - check for vm changes every 1 second (salt default)
    # - does not send events at startup
    beacons:
      vmadm: []

    ## standard
    # - check for vm changes every 60 seconds
    # - send create event at startup for all vms
    beacons:
      vmadm:
        - interval: 60
        - startup_create_event: True
"""

import logging

import salt.utils.beacons

__virtualname__ = "vmadm"

VMADM_STATE = {
    "first_run": True,
    "vms": [],
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Provides vmadm beacon on SmartOS
    """
    if "vmadm.list" in __salt__:
        return True
    else:
        err_msg = "Only available on SmartOS compute nodes."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg


def validate(config):
    """
    Validate the beacon configuration
    """
    vcfg_ret = True
    vcfg_msg = "Valid beacon configuration"

    if not isinstance(config, list):
        vcfg_ret = False
        vcfg_msg = "Configuration for vmadm beacon must be a list!"

    return vcfg_ret, vcfg_msg


def beacon(config):
    """
    Poll vmadm for changes
    """
    ret = []

    # NOTE: lookup current images
    current_vms = __salt__["vmadm.list"](
        keyed=True,
        order="uuid,state,alias,hostname,dns_domain",
    )

    # NOTE: apply configuration
    if VMADM_STATE["first_run"]:
        log.info("Applying configuration for vmadm beacon")

        config = salt.utils.beacons.list_to_dict(config)

        if "startup_create_event" not in config or not config["startup_create_event"]:
            VMADM_STATE["vms"] = current_vms

    # NOTE: create events
    for uuid in current_vms:
        event = {}
        if uuid not in VMADM_STATE["vms"]:
            event["tag"] = f"created/{uuid}"
            for label in current_vms[uuid]:
                if label == "state":
                    continue
                event[label] = current_vms[uuid][label]

        if event:
            ret.append(event)

    # NOTE: deleted events
    for uuid in VMADM_STATE["vms"]:
        event = {}
        if uuid not in current_vms:
            event["tag"] = f"deleted/{uuid}"
            for label in VMADM_STATE["vms"][uuid]:
                if label == "state":
                    continue
                event[label] = VMADM_STATE["vms"][uuid][label]

        if event:
            ret.append(event)

    # NOTE: state change events
    for uuid in current_vms:
        event = {}
        if (
            VMADM_STATE["first_run"]
            or uuid not in VMADM_STATE["vms"]
            or current_vms[uuid].get("state", "unknown")
            != VMADM_STATE["vms"][uuid].get("state", "unknown")
        ):
            event["tag"] = "{}/{}".format(
                current_vms[uuid].get("state", "unknown"), uuid
            )
            for label in current_vms[uuid]:
                if label == "state":
                    continue
                event[label] = current_vms[uuid][label]

        if event:
            ret.append(event)

    # NOTE: update stored state
    VMADM_STATE["vms"] = current_vms

    # NOTE: disable first_run
    if VMADM_STATE["first_run"]:
        VMADM_STATE["first_run"] = False

    return ret



"""
Beacon to monitor memory usage.

.. versionadded:: 2016.3.0

:depends: python-psutil
"""

import logging
import re

import salt.utils.beacons

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

log = logging.getLogger(__name__)

__virtualname__ = "memusage"


def __virtual__():
    if HAS_PSUTIL is False:
        err_msg = "psutil library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg
    else:
        return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for memusage beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for memusage beacon must be a list."
    else:
        config = salt.utils.beacons.list_to_dict(config)

        if "percent" not in config:
            return False, "Configuration for memusage beacon requires percent."
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Monitor the memory usage of the minion

    Specify thresholds for percent used and only emit a beacon
    if it is exceeded.

    .. code-block:: yaml

        beacons:
          memusage:
            - percent: 63%
    """
    ret = []

    config = salt.utils.beacons.list_to_dict(config)

    _current_usage = psutil.virtual_memory()

    current_usage = _current_usage.percent
    monitor_usage = config["percent"]
    if isinstance(monitor_usage, str) and "%" in monitor_usage:
        monitor_usage = re.sub("%", "", monitor_usage)
    monitor_usage = float(monitor_usage)
    if current_usage >= monitor_usage:
        ret.append({"memusage": current_usage})
    return ret
